# Copyright (c) 2016 Juniper Networks Inc.
# All Rights Reserved.
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

import copy
import operator
from unittest import mock

from osc_lib import exceptions
from osc_lib import utils as osc_utils
from osc_lib.utils import columns as column_util

from openstackclient.network.v2.bgpvpn import bgpvpn
from openstackclient.tests.unit.network.v2.bgpvpn import fakes


columns_short = tuple(
    col
    for col, _, listing_mode in bgpvpn._attr_map
    if listing_mode in (column_util.LIST_BOTH, column_util.LIST_SHORT_ONLY)
)
columns_long = tuple(
    col
    for col, _, listing_mode in bgpvpn._attr_map
    if listing_mode in (column_util.LIST_BOTH, column_util.LIST_LONG_ONLY)
)
headers_short = tuple(
    head
    for _, head, listing_mode in bgpvpn._attr_map
    if listing_mode in (column_util.LIST_BOTH, column_util.LIST_SHORT_ONLY)
)
headers_long = tuple(
    head
    for _, head, listing_mode in bgpvpn._attr_map
    if listing_mode in (column_util.LIST_BOTH, column_util.LIST_LONG_ONLY)
)
sorted_attr_map = sorted(bgpvpn._attr_map, key=operator.itemgetter(1))
sorted_columns = tuple(col for col, _, _ in sorted_attr_map)
sorted_headers = tuple(head for _, head, _ in sorted_attr_map)


def _get_data(attrs, columns=sorted_columns):
    return osc_utils.get_dict_properties(
        attrs, columns, formatters=bgpvpn._formatters
    )


class TestCreateBgpvpn(fakes.TestNeutronClientBgpvpn):
    def setUp(self):
        super().setUp()
        self.cmd = bgpvpn.CreateBgpvpn(self.app, None)

    def test_create_bgpvpn_with_no_args(self):
        fake_bgpvpn = fakes.create_one_bgpvpn()
        self.network_client.create_bgpvpn.return_value = fake_bgpvpn
        arglist = []
        verifylist = [
            ('project', None),
            ('name', None),
            ('type', 'l3'),
            ('vni', None),
            ('local_pref', None),
            ('route_targets', None),
            ('import_targets', None),
            ('export_targets', None),
            ('route_distinguishers', None),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        cols, _data = self.cmd.take_action(parsed_args)

        self.network_client.create_bgpvpn.assert_called_once_with(
            **{'type': 'l3'}
        )

        self.assertEqual(sorted(sorted_columns), sorted(cols))

    @mock.patch('osc_lib.cli.identity.find_project')
    def test_create_bgpvpn_with_all_args(self, mock_find_project):
        attrs = {
            'project_id': 'new_fake_project_id',
            'name': 'fake_name',
            'type': 'l2',
            'vni': 100,
            'local_pref': 777,
            'route_targets': ['fake_rt1', 'fake_rt2', 'fake_rt3'],
            'import_targets': ['fake_irt1', 'fake_irt2', 'fake_irt3'],
            'export_targets': ['fake_ert1', 'fake_ert2', 'fake_ert3'],
            'route_distinguishers': ['fake_rd1', 'fake_rd2', 'fake_rd3'],
        }
        fake_bgpvpn = fakes.create_one_bgpvpn(attrs)
        self.network_client.create_bgpvpn.return_value = fake_bgpvpn
        mock_find_project.return_value = mock.Mock(
            id=fake_bgpvpn['project_id']
        )
        arglist = [
            '--project',
            fake_bgpvpn['project_id'],
            '--name',
            fake_bgpvpn['name'],
            '--type',
            fake_bgpvpn['type'],
            '--vni',
            str(fake_bgpvpn['vni']),
            '--local-pref',
            str(fake_bgpvpn['local_pref']),
        ]
        for rt in fake_bgpvpn['route_targets']:
            arglist.extend(['--route-target', rt])
        for rt in fake_bgpvpn['import_targets']:
            arglist.extend(['--import-target', rt])
        for rt in fake_bgpvpn['export_targets']:
            arglist.extend(['--export-target', rt])
        for rd in fake_bgpvpn['route_distinguishers']:
            arglist.extend(['--route-distinguisher', rd])
        verifylist = [
            ('project', fake_bgpvpn['project_id']),
            ('name', fake_bgpvpn['name']),
            ('type', fake_bgpvpn['type']),
            ('vni', fake_bgpvpn['vni']),
            ('local_pref', fake_bgpvpn['local_pref']),
            ('route_targets', fake_bgpvpn['route_targets']),
            ('import_targets', fake_bgpvpn['import_targets']),
            ('export_targets', fake_bgpvpn['export_targets']),
            ('route_distinguishers', fake_bgpvpn['route_distinguishers']),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        cols, _data = self.cmd.take_action(parsed_args)

        fake_bgpvpn_call = copy.deepcopy(attrs)

        self.network_client.create_bgpvpn.assert_called_once_with(
            **fake_bgpvpn_call
        )
        self.assertEqual(sorted(sorted_columns), sorted(cols))


class TestSetBgpvpn(fakes.TestNeutronClientBgpvpn):
    def setUp(self):
        super().setUp()
        self.network_client.find_bgpvpn.side_effect = (
            lambda name_or_id, **kwargs: {'id': name_or_id}
        )

        self.cmd = bgpvpn.SetBgpvpn(self.app, None)

    def test_set_bgpvpn(self):
        attrs = {
            'route_targets': ['set_rt1', 'set_rt2', 'set_rt3'],
            'import_targets': ['set_irt1', 'set_irt2', 'set_irt3'],
            'export_targets': ['set_ert1', 'set_ert2', 'set_ert3'],
            'route_distinguishers': ['set_rd1', 'set_rd2', 'set_rd3'],
        }
        fake_bgpvpn = fakes.create_one_bgpvpn(attrs)
        self.network_client.get_bgpvpn.return_value = fake_bgpvpn
        arglist = [
            fake_bgpvpn['id'],
            '--name',
            'set_name',
            '--route-target',
            'set_rt1',
            '--import-target',
            'set_irt1',
            '--export-target',
            'set_ert1',
            '--route-distinguisher',
            'set_rd1',
        ]
        verifylist = [
            ('bgpvpn', fake_bgpvpn['id']),
            ('name', 'set_name'),
            ('route_targets', ['set_rt1']),
            ('purge_route_target', False),
            ('import_targets', ['set_irt1']),
            ('purge_import_target', False),
            ('export_targets', ['set_ert1']),
            ('purge_export_target', False),
            ('route_distinguishers', ['set_rd1']),
            ('purge_route_distinguisher', False),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        attrs = {
            'name': 'set_name',
            'route_targets': list(
                set(fake_bgpvpn['route_targets']) | set(['set_rt1'])
            ),
            'import_targets': list(
                set(fake_bgpvpn['import_targets']) | set(['set_irt1'])
            ),
            'export_targets': list(
                set(fake_bgpvpn['export_targets']) | set(['set_ert1'])
            ),
            'route_distinguishers': list(
                set(fake_bgpvpn['route_distinguishers']) | set(['set_rd1'])
            ),
        }
        self.network_client.update_bgpvpn.assert_called_once_with(
            fake_bgpvpn['id'], **attrs
        )
        self.assertIsNone(result)

    def test_set_bgpvpn_with_purge_list(self):
        fake_bgpvpn = fakes.create_one_bgpvpn()
        self.network_client.get_bgpvpn.return_value = fake_bgpvpn
        arglist = [
            fake_bgpvpn['id'],
            '--route-target',
            'set_rt1',
            '--no-route-target',
            '--import-target',
            'set_irt1',
            '--no-import-target',
            '--export-target',
            'set_ert1',
            '--no-export-target',
            '--route-distinguisher',
            'set_rd1',
            '--no-route-distinguisher',
        ]
        verifylist = [
            ('bgpvpn', fake_bgpvpn['id']),
            ('route_targets', ['set_rt1']),
            ('purge_route_target', True),
            ('import_targets', ['set_irt1']),
            ('purge_import_target', True),
            ('export_targets', ['set_ert1']),
            ('purge_export_target', True),
            ('route_distinguishers', ['set_rd1']),
            ('purge_route_distinguisher', True),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        attrs = {
            'route_targets': [],
            'import_targets': [],
            'export_targets': [],
            'route_distinguishers': [],
        }
        self.network_client.update_bgpvpn.assert_called_once_with(
            fake_bgpvpn['id'], **attrs
        )
        self.assertIsNone(result)


class TestUnsetBgpvpn(fakes.TestNeutronClientBgpvpn):
    def setUp(self):
        super().setUp()
        self.network_client.find_bgpvpn.side_effect = (
            lambda name_or_id, **kwargs: {'id': name_or_id}
        )
        self.cmd = bgpvpn.UnsetBgpvpn(self.app, None)

    def test_unset_bgpvpn(self):
        attrs = {
            'route_targets': ['unset_rt1', 'unset_rt2', 'unset_rt3'],
            'import_targets': ['unset_irt1', 'unset_irt2', 'unset_irt3'],
            'export_targets': ['unset_ert1', 'unset_ert2', 'unset_ert3'],
            'route_distinguishers': ['unset_rd1', 'unset_rd2', 'unset_rd3'],
        }
        fake_bgpvpn = fakes.create_one_bgpvpn(attrs)
        self.network_client.get_bgpvpn.return_value = fake_bgpvpn
        arglist = [
            fake_bgpvpn['id'],
            '--route-target',
            'unset_rt1',
            '--import-target',
            'unset_irt1',
            '--export-target',
            'unset_ert1',
            '--route-distinguisher',
            'unset_rd1',
        ]
        verifylist = [
            ('bgpvpn', fake_bgpvpn['id']),
            ('route_targets', ['unset_rt1']),
            ('purge_route_target', False),
            ('import_targets', ['unset_irt1']),
            ('purge_import_target', False),
            ('export_targets', ['unset_ert1']),
            ('purge_export_target', False),
            ('route_distinguishers', ['unset_rd1']),
            ('purge_route_distinguisher', False),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        attrs = {
            'route_targets': list(
                set(fake_bgpvpn['route_targets']) - set(['unset_rt1'])
            ),
            'import_targets': list(
                set(fake_bgpvpn['import_targets']) - set(['unset_irt1'])
            ),
            'export_targets': list(
                set(fake_bgpvpn['export_targets']) - set(['unset_ert1'])
            ),
            'route_distinguishers': list(
                set(fake_bgpvpn['route_distinguishers']) - set(['unset_rd1'])
            ),
        }
        self.network_client.update_bgpvpn.assert_called_once_with(
            fake_bgpvpn['id'], **attrs
        )
        self.assertIsNone(result)

    def test_unset_bgpvpn_with_purge_list(self):
        fake_bgpvpn = fakes.create_one_bgpvpn()
        self.network_client.get_bgpvpn.return_value = fake_bgpvpn
        arglist = [
            fake_bgpvpn['id'],
            '--route-target',
            'unset_rt1',
            '--all-route-target',
            '--import-target',
            'unset_irt1',
            '--all-import-target',
            '--export-target',
            'unset_ert1',
            '--all-export-target',
            '--route-distinguisher',
            'unset_rd1',
            '--all-route-distinguisher',
        ]
        verifylist = [
            ('bgpvpn', fake_bgpvpn['id']),
            ('route_targets', ['unset_rt1']),
            ('purge_route_target', True),
            ('import_targets', ['unset_irt1']),
            ('purge_import_target', True),
            ('export_targets', ['unset_ert1']),
            ('purge_export_target', True),
            ('route_distinguishers', ['unset_rd1']),
            ('purge_route_distinguisher', True),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        attrs = {
            'route_targets': [],
            'import_targets': [],
            'export_targets': [],
            'route_distinguishers': [],
        }
        self.network_client.update_bgpvpn.assert_called_once_with(
            fake_bgpvpn['id'], **attrs
        )
        self.assertIsNone(result)


class TestDeleteBgpvpn(fakes.TestNeutronClientBgpvpn):
    def setUp(self):
        super().setUp()
        self.network_client.find_bgpvpn.side_effect = (
            lambda name_or_id, **kwargs: {'id': name_or_id}
        )
        self.cmd = bgpvpn.DeleteBgpvpn(self.app, None)

    def test_delete_one_bgpvpn(self):
        fake_bgpvpn = fakes.create_one_bgpvpn()
        arglist = [
            fake_bgpvpn['id'],
        ]
        verifylist = [
            ('bgpvpns', [fake_bgpvpn['id']]),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        self.network_client.delete_bgpvpn.assert_called_once_with(
            fake_bgpvpn['id']
        )
        self.assertIsNone(result)

    def test_delete_multi_bpgvpn(self):
        fake_bgpvpns = fakes.create_bgpvpns(count=3)
        fake_bgpvpn_ids = [fake_bgpvpn['id'] for fake_bgpvpn in fake_bgpvpns]
        arglist = fake_bgpvpn_ids
        verifylist = [
            ('bgpvpns', fake_bgpvpn_ids),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        self.network_client.delete_bgpvpn.assert_has_calls(
            [mock.call(id) for id in fake_bgpvpn_ids]
        )
        self.assertIsNone(result)

    def test_delete_multi_bpgvpn_with_unknown(self):
        count = 3
        fake_bgpvpns = fakes.create_bgpvpns(count=count)
        fake_bgpvpn_ids = [fake_bgpvpn['id'] for fake_bgpvpn in fake_bgpvpns]

        def raise_unknonw_resource(resource_path, name_or_id):
            if str(count - 2) in name_or_id:
                raise Exception()

        self.network_client.delete_bgpvpn.side_effect = raise_unknonw_resource
        arglist = fake_bgpvpn_ids
        verifylist = [
            ('bgpvpns', fake_bgpvpn_ids),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        self.assertRaises(
            exceptions.CommandError, self.cmd.take_action, parsed_args
        )

        self.network_client.delete_bgpvpn.assert_has_calls(
            [mock.call(id) for id in fake_bgpvpn_ids]
        )


class TestListBgpvpn(fakes.TestNeutronClientBgpvpn):
    def setUp(self):
        super().setUp()
        self.cmd = bgpvpn.ListBgpvpn(self.app, None)

    def test_list_all_bgpvpn(self):
        count = 3
        fake_bgpvpns = fakes.create_bgpvpns(count=count)
        self.network_client.bgpvpns.return_value = fake_bgpvpns
        arglist = []
        verifylist = []

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        headers, data = self.cmd.take_action(parsed_args)

        self.network_client.bgpvpns.assert_called_once()
        self.assertEqual(headers, list(headers_short))
        self.assertListEqual(
            list(data),
            [
                _get_data(fake_bgpvpn, columns_short)
                for fake_bgpvpn in fake_bgpvpns
            ],
        )

    def test_list_all_bgpvpn_long_mode(self):
        count = 3
        fake_bgpvpns = fakes.create_bgpvpns(count=count)
        self.network_client.bgpvpns.return_value = fake_bgpvpns
        arglist = [
            '--long',
        ]
        verifylist = [
            ('long', True),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        headers, data = self.cmd.take_action(parsed_args)

        self.network_client.bgpvpns.assert_called_once()
        self.assertEqual(headers, list(headers_long))
        self.assertListEqual(
            list(data),
            [
                _get_data(fake_bgpvpn, columns_long)
                for fake_bgpvpn in fake_bgpvpns
            ],
        )

    @mock.patch('osc_lib.cli.identity.find_project')
    def test_list_project_bgpvpn(self, mock_find_project):
        count = 3
        project_id = 'list_fake_project_id'
        attrs = {'project_id': project_id}
        fake_bgpvpns = fakes.create_bgpvpns(count=count, attrs=attrs)
        self.network_client.bgpvpns.return_value = fake_bgpvpns
        mock_find_project.return_value = mock.Mock(id=project_id)
        arglist = [
            '--project',
            project_id,
        ]
        verifylist = [
            ('project', project_id),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        headers, data = self.cmd.take_action(parsed_args)

        self.network_client.bgpvpns.assert_called_once_with(
            project_id=project_id
        )
        self.assertEqual(headers, list(headers_short))
        self.assertListEqual(
            list(data),
            [
                _get_data(fake_bgpvpn, columns_short)
                for fake_bgpvpn in fake_bgpvpns
            ],
        )

    def test_list_bgpvpn_with_filters(self):
        count = 3
        name = 'fake_id0'
        layer_type = 'l2'
        attrs = {'type': layer_type}
        fake_bgpvpns = fakes.create_bgpvpns(count=count, attrs=attrs)
        returned_bgpvpn = fake_bgpvpns[0]
        self.network_client.bgpvpns.return_value = [returned_bgpvpn]
        arglist = [
            '--property',
            f'name={name}',
            '--property',
            f'type={layer_type}',
        ]
        verifylist = [
            ('property', {'name': name, 'type': layer_type}),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        headers, data = self.cmd.take_action(parsed_args)

        self.network_client.bgpvpns.assert_called_once_with(
            name=name, type=layer_type
        )
        self.assertEqual(headers, list(headers_short))
        self.assertListEqual(
            list(data), [_get_data(returned_bgpvpn, columns_short)]
        )


class TestShowBgpvpn(fakes.TestNeutronClientBgpvpn):
    def setUp(self):
        super().setUp()
        self.cmd = bgpvpn.ShowBgpvpn(self.app, None)
        self.network_client.find_bgpvpn.side_effect = (
            lambda name_or_id, **kwargs: {'id': name_or_id}
        )

    def test_show_bgpvpn(self):
        fake_bgpvpn = fakes.create_one_bgpvpn()
        self.network_client.get_bgpvpn.return_value = fake_bgpvpn
        arglist = [
            fake_bgpvpn['id'],
        ]
        verifylist = [
            ('bgpvpn', fake_bgpvpn['id']),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        headers, _data = self.cmd.take_action(parsed_args)

        self.network_client.get_bgpvpn.assert_called_once_with(
            fake_bgpvpn['id']
        )
        self.assertEqual(sorted(sorted_columns), sorted(headers))



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""Compute v2 Server Migration action implementations"""

import argparse
from collections.abc import Iterable, Sequence
from typing import Any
import uuid

from openstack import utils as sdk_utils
from osc_lib import exceptions
from osc_lib import utils

from openstackclient import command
from openstackclient.common import pagination
from openstackclient.i18n import _
from openstackclient.identity import common as identity_common


class ListMigration(command.Lister):
    _description = _("""List server migrations""")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            '--server',
            metavar='<server>',
            help=_('Filter migrations by server (name or ID)'),
        )
        parser.add_argument(
            '--host',
            metavar='<host>',
            help=_('Filter migrations by source or destination host'),
        )
        parser.add_argument(
            '--status',
            metavar='<status>',
            help=_('Filter migrations by status'),
        )
        parser.add_argument(
            '--type',
            metavar='<type>',
            choices=[
                'evacuation',
                'live-migration',
                'cold-migration',
                'resize',
            ],
            help=_('Filter migrations by type'),
        )
        pagination.add_marker_pagination_option_to_parser(parser)
        parser.add_argument(
            '--changes-since',
            dest='changes_since',
            metavar='<changes-since>',
            help=_(
                "List only migrations changed later or equal to a certain "
                "point of time. The provided time should be an ISO 8061 "
                "formatted time, e.g. ``2016-03-04T06:27:59Z``. "
                "(supported with --os-compute-api-version 2.59 or above)"
            ),
        )
        parser.add_argument(
            '--changes-before',
            dest='changes_before',
            metavar='<changes-before>',
            help=_(
                "List only migrations changed earlier or equal to a "
                "certain point of time. The provided time should be an ISO "
                "8061 formatted time, e.g. ``2016-03-04T06:27:59Z``. "
                "(supported with --os-compute-api-version 2.66 or above)"
            ),
        )
        parser.add_argument(
            '--project',
            metavar='<project>',
            help=_(
                "Filter migrations by project (name or ID) "
                "(supported with --os-compute-api-version 2.80 or above)"
            ),
        )
        identity_common.add_project_domain_option_to_parser(parser)
        parser.add_argument(
            '--user',
            metavar='<user>',
            help=_(
                "Filter migrations by user (name or ID) "
                "(supported with --os-compute-api-version 2.80 or above)"
            ),
        )
        identity_common.add_user_domain_option_to_parser(parser)
        return parser

    def print_migrations(
        self,
        parsed_args: argparse.Namespace,
        compute_client: Any,
        migrations: Any,
    ) -> tuple[list[str], Iterable[tuple[Any, ...]]]:
        column_headers = [
            'Source Node',
            'Dest Node',
            'Source Compute',
            'Dest Compute',
            'Dest Host',
            'Status',
            'Server UUID',
            'Old Flavor',
            'New Flavor',
            'Created At',
            'Updated At',
        ]

        # Response fields coming back from the REST API are not always exactly
        # the same as the column header names.
        columns = [
            'source_node',
            'dest_node',
            'source_compute',
            'dest_compute',
            'dest_host',
            'status',
            'server_id',
            'old_flavor_id',
            'new_flavor_id',
            'created_at',
            'updated_at',
        ]

        # Insert migrations UUID after ID
        if sdk_utils.supports_microversion(compute_client, "2.59"):
            column_headers.insert(0, "UUID")
            columns.insert(0, "uuid")

        if sdk_utils.supports_microversion(compute_client, "2.23"):
            column_headers.insert(0, "Id")
            columns.insert(0, "id")
            column_headers.insert(len(column_headers) - 2, "Type")
            columns.insert(len(columns) - 2, "migration_type")

        if sdk_utils.supports_microversion(compute_client, "2.80"):
            if parsed_args.project:
                column_headers.insert(len(column_headers) - 2, "Project")
                columns.insert(len(columns) - 2, "project_id")
            if parsed_args.user:
                column_headers.insert(len(column_headers) - 2, "User")
                columns.insert(len(columns) - 2, "user_id")

        return (
            column_headers,
            (utils.get_item_properties(mig, columns) for mig in migrations),
        )

    def take_action(
        self, parsed_args: argparse.Namespace
    ) -> tuple[list[str], Iterable[tuple[Any, ...]]]:
        compute_client = self.app.client_manager.compute
        identity_client = self.app.client_manager.identity

        search_opts = {}

        if parsed_args.host is not None:
            search_opts['host'] = parsed_args.host

        if parsed_args.status is not None:
            search_opts['status'] = parsed_args.status

        if parsed_args.server:
            server = compute_client.find_server(
                parsed_args.server, ignore_missing=False
            )
            search_opts['instance_uuid'] = server.id

        if parsed_args.type:
            migration_type = parsed_args.type
            # we're using an alias because the default value is confusing
            if migration_type == 'cold-migration':
                migration_type = 'migration'
            search_opts['migration_type'] = migration_type

        if parsed_args.marker:
            if not sdk_utils.supports_microversion(compute_client, "2.59"):
                msg = _(
                    '--os-compute-api-version 2.59 or greater is required to '
                    'support the --marker option'
                )
                raise exceptions.CommandError(msg)
            search_opts['marker'] = parsed_args.marker

        if parsed_args.limit:
            if not sdk_utils.supports_microversion(compute_client, "2.59"):
                msg = _(
                    '--os-compute-api-version 2.59 or greater is required to '
                    'support the --limit option'
                )
                raise exceptions.CommandError(msg)
            search_opts['limit'] = parsed_args.limit
            search_opts['paginated'] = False

        if parsed_args.changes_since:
            if not sdk_utils.supports_microversion(compute_client, "2.59"):
                msg = _(
                    '--os-compute-api-version 2.59 or greater is required to '
                    'support the --changes-since option'
                )
                raise exceptions.CommandError(msg)
            search_opts['changes_since'] = parsed_args.changes_since

        if parsed_args.changes_before:
            if not sdk_utils.supports_microversion(compute_client, "2.66"):
                msg = _(
                    '--os-compute-api-version 2.66 or greater is required to '
                    'support the --changes-before option'
                )
                raise exceptions.CommandError(msg)
            search_opts['changes_before'] = parsed_args.changes_before

        if parsed_args.project:
            if not sdk_utils.supports_microversion(compute_client, "2.80"):
                msg = _(
                    '--os-compute-api-version 2.80 or greater is required to '
                    'support the --project option'
                )
                raise exceptions.CommandError(msg)

            search_opts['project_id'] = identity_common.find_project(
                identity_client,
                parsed_args.project,
                parsed_args.project_domain,
            ).id

        if parsed_args.user:
            if not sdk_utils.supports_microversion(compute_client, "2.80"):
                msg = _(
                    '--os-compute-api-version 2.80 or greater is required to '
                    'support the --user option'
                )
                raise exceptions.CommandError(msg)

            search_opts['user_id'] = identity_common.find_user(
                identity_client,
                parsed_args.user,
                parsed_args.user_domain,
            ).id

        migrations = list(compute_client.migrations(**search_opts))

        return self.print_migrations(parsed_args, compute_client, migrations)


def _get_migration_by_uuid(
    compute_client: Any, server_id: str, migration_uuid: str
) -> Any:
    for migration in compute_client.server_migrations(server_id):
        if migration.uuid == migration_uuid:
            return migration
    else:
        msg = _(
            'In-progress live migration %(migration)s is not found for '
            'server %(server)s'
        )
        raise exceptions.CommandError(
            msg
            % {
                'migration': migration_uuid,
                'server': server_id,
            }
        )


class ShowMigration(command.ShowOne):
    """Show an in-progress live migration for a given server.

    Note that it is not possible to show cold migrations or completed
    live-migrations. Use 'openstack server migration list' to get details for
    these.
    """

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'server',
            metavar='<server>',
            help=_('Server (name or ID)'),
        )
        parser.add_argument(
            'migration',
            metavar='<migration>',
            help=_("Migration (ID)"),
        )
        return parser

    def take_action(
        self, parsed_args: argparse.Namespace
    ) -> tuple[Sequence[str], Iterable[Any]]:
        compute_client = self.app.client_manager.compute

        if not sdk_utils.supports_microversion(compute_client, '2.24'):
            msg = _(
                '--os-compute-api-version 2.24 or greater is required to '
                'support the server migration show command'
            )
            raise exceptions.CommandError(msg)

        if not parsed_args.migration.isdigit():
            try:
                uuid.UUID(parsed_args.migration)
            except ValueError:
                msg = _('The <migration> argument must be an ID or UUID')
                raise exceptions.CommandError(msg)

            if not sdk_utils.supports_microversion(compute_client, '2.59'):
                msg = _(
                    '--os-compute-api-version 2.59 or greater is required to '
                    'retrieve server migrations by UUID'
                )
                raise exceptions.CommandError(msg)

        server = compute_client.find_server(
            parsed_args.server,
            ignore_missing=False,
        )

        # the nova API doesn't currently allow retrieval by UUID but it's a
        # reasonably common operation so emulate this behavior by listing
        # migrations - the responses are identical
        if not parsed_args.migration.isdigit():
            server_migration = _get_migration_by_uuid(
                compute_client,
                server.id,
                parsed_args.migration,
            )
        else:
            server_migration = compute_client.get_server_migration(
                server.id,
                parsed_args.migration,
                ignore_missing=False,
            )

        column_headers: tuple[str, ...] = (
            'ID',
            'Server UUID',
            'Status',
            'Source Compute',
            'Source Node',
            'Dest Compute',
            'Dest Host',
            'Dest Node',
            'Memory Total Bytes',
            'Memory Processed Bytes',
            'Memory Remaining Bytes',
            'Disk Total Bytes',
            'Disk Processed Bytes',
            'Disk Remaining Bytes',
            'Created At',
            'Updated At',
        )

        columns: tuple[str, ...] = (
            'id',
            'server_id',
            'status',
            'source_compute',
            'source_node',
            'dest_compute',
            'dest_host',
            'dest_node',
            'memory_total_bytes',
            'memory_processed_bytes',
            'memory_remaining_bytes',
            'disk_total_bytes',
            'disk_processed_bytes',
            'disk_remaining_bytes',
            'created_at',
            'updated_at',
        )

        if sdk_utils.supports_microversion(compute_client, '2.59'):
            column_headers += ('UUID',)
            columns += ('uuid',)

        if sdk_utils.supports_microversion(compute_client, '2.80'):
            column_headers += ('User ID', 'Project ID')
            columns += ('user_id', 'project_id')

        data = utils.get_item_properties(server_migration, columns)
        return column_headers, data


class AbortMigration(command.Command):
    """Cancel an ongoing live migration.

    This command requires ``--os-compute-api-version`` 2.24 or greater.
    """

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'server',
            metavar='<server>',
            help=_('Server (name or ID)'),
        )
        parser.add_argument(
            'migration',
            metavar='<migration>',
            help=_("Migration (ID)"),
        )
        return parser

    def take_action(self, parsed_args: argparse.Namespace) -> None:
        compute_client = self.app.client_manager.compute

        if not sdk_utils.supports_microversion(compute_client, '2.24'):
            msg = _(
                '--os-compute-api-version 2.24 or greater is required to '
                'support the server migration abort command'
            )
            raise exceptions.CommandError(msg)

        if not parsed_args.migration.isdigit():
            try:
                uuid.UUID(parsed_args.migration)
            except ValueError:
                msg = _('The <migration> argument must be an ID or UUID')
                raise exceptions.CommandError(msg)

            if not sdk_utils.supports_microversion(compute_client, '2.59'):
                msg = _(
                    '--os-compute-api-version 2.59 or greater is required to '
                    'abort server migrations by UUID'
                )
                raise exceptions.CommandError(msg)

        server = compute_client.find_server(
            parsed_args.server,
            ignore_missing=False,
        )

        # the nova API doesn't currently allow retrieval by UUID but it's a
        # reasonably common operation so emulate this behavior by listing
        # migrations - the responses are identical
        migration_id = parsed_args.migration
        if not parsed_args.migration.isdigit():
            migration_id = _get_migration_by_uuid(
                compute_client,
                server.id,
                parsed_args.migration,
            ).id

        compute_client.abort_server_migration(
            migration_id,
            server.id,
            ignore_missing=False,
        )


class ForceCompleteMigration(command.Command):
    """Force an ongoing live migration to complete.

    This command requires ``--os-compute-api-version`` 2.22 or greater.
    """

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'server',
            metavar='<server>',
            help=_('Server (name or ID)'),
        )
        parser.add_argument(
            'migration', metavar='<migration>', help=_('Migration (ID)')
        )
        return parser

    def take_action(self, parsed_args: argparse.Namespace) -> None:
        compute_client = self.app.client_manager.compute

        if not sdk_utils.supports_microversion(compute_client, '2.22'):
            msg = _(
                '--os-compute-api-version 2.22 or greater is required to '
                'support the server migration force complete command'
            )
            raise exceptions.CommandError(msg)

        if not parsed_args.migration.isdigit():
            try:
                uuid.UUID(parsed_args.migration)
            except ValueError:
                msg = _('The <migration> argument must be an ID or UUID')
                raise exceptions.CommandError(msg)

            if not sdk_utils.supports_microversion(compute_client, '2.59'):
                msg = _(
                    '--os-compute-api-version 2.59 or greater is required to '
                    'abort server migrations by UUID'
                )
                raise exceptions.CommandError(msg)

        server = compute_client.find_server(
            parsed_args.server,
            ignore_missing=False,
        )

        # the nova API doesn't currently allow retrieval by UUID but it's a
        # reasonably common operation so emulate this behavior by listing
        # migrations - the responses are identical
        migration_id = parsed_args.migration
        if not parsed_args.migration.isdigit():
            migration_id = _get_migration_by_uuid(
                compute_client,
                server.id,
                parsed_args.migration,
            ).id

        compute_client.force_complete_server_migration(migration_id, server.id)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from tempest.lib.common.utils import data_utils

from openstackclient.tests.functional.identity.v3 import common


class GroupTests(common.IdentityTests):
    def test_group_create(self):
        self._create_dummy_group()

    def test_group_list(self):
        group_name = self._create_dummy_group()
        raw_output = self.openstack('group list')
        items = self.parse_listing(raw_output)
        self.assert_table_structure(items, common.BASIC_LIST_HEADERS)
        self.assertIn(group_name, raw_output)

    def test_group_list_with_domain(self):
        group_name = self._create_dummy_group()
        raw_output = self.openstack(f'group list --domain {self.DOMAIN_NAME}')
        items = self.parse_listing(raw_output)
        self.assert_table_structure(items, common.BASIC_LIST_HEADERS)
        self.assertIn(group_name, raw_output)

    def test_group_delete(self):
        group_name = self._create_dummy_group(add_clean_up=False)
        raw_output = self.openstack(
            f'group delete --domain {self.DOMAIN_NAME} {group_name}'
        )
        self.assertEqual(0, len(raw_output))

    def test_group_show(self):
        group_name = self._create_dummy_group()
        raw_output = self.openstack(
            f'group show --domain {self.DOMAIN_NAME} {group_name}'
        )
        items = self.parse_show(raw_output)
        self.assert_show_fields(items, self.GROUP_FIELDS)

    def test_group_set(self):
        group_name = self._create_dummy_group()
        new_group_name = data_utils.rand_name('NewTestGroup')
        raw_output = self.openstack(
            'group set '
            f'--domain {self.DOMAIN_NAME} '
            f'--name {new_group_name} '
            f'{group_name}'
        )
        self.assertEqual(0, len(raw_output))
        raw_output = self.openstack(
            f'group show --domain {self.DOMAIN_NAME} {new_group_name}'
        )
        group = self.parse_show_as_object(raw_output)
        self.assertEqual(new_group_name, group['name'])
        # reset group name to make sure it will be cleaned up
        raw_output = self.openstack(
            'group set '
            f'--domain {self.DOMAIN_NAME} '
            f'--name {group_name} '
            f'{new_group_name}'
        )
        self.assertEqual(0, len(raw_output))

    def test_group_add_user(self):
        group_name = self._create_dummy_group()
        username = self._create_dummy_user()
        raw_output = self.openstack(
            'group add user '
            f'--group-domain {self.DOMAIN_NAME} '
            f'--user-domain {self.DOMAIN_NAME} '
            f'{group_name} {username}'
        )
        self.addCleanup(
            self.openstack,
            'group remove user '
            f'--group-domain {self.DOMAIN_NAME} '
            f'--user-domain {self.DOMAIN_NAME} '
            f'{group_name} {username}',
        )
        self.assertOutput('', raw_output)

    def test_group_contains_user(self):
        group_name = self._create_dummy_group()
        username = self._create_dummy_user()
        raw_output = self.openstack(
            'group add user '
            f'--group-domain {self.DOMAIN_NAME} '
            f'--user-domain {self.DOMAIN_NAME} '
            f'{group_name} {username}'
        )
        self.addCleanup(
            self.openstack,
            'group remove user '
            f'--group-domain {self.DOMAIN_NAME} '
            f'--user-domain {self.DOMAIN_NAME} '
            f'{group_name} {username}',
        )
        self.assertOutput('', raw_output)
        raw_output = self.openstack(
            'group contains user '
            f'--group-domain {self.DOMAIN_NAME} '
            f'--user-domain {self.DOMAIN_NAME} '
            f'{group_name} {username}'
        )
        self.assertEqual(
            f'{username} in group {group_name}\n',
            raw_output,
        )

    def test_group_remove_user(self):
        group_name = self._create_dummy_group()
        username = self._create_dummy_user()
        add_raw_output = self.openstack(
            'group add user '
            f'--group-domain {self.DOMAIN_NAME} '
            f'--user-domain {self.DOMAIN_NAME} '
            f'{group_name} {username}'
        )
        remove_raw_output = self.openstack(
            'group remove user '
            f'--group-domain {self.DOMAIN_NAME} '
            f'--user-domain {self.DOMAIN_NAME} '
            f'{group_name} {username}'
        )
        self.assertOutput('', add_raw_output)
        self.assertOutput('', remove_raw_output)



# Copyright (c) 2016 Juniper Networks Inc.
# All Rights Reserved.
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

import logging
import typing as ty

from osc_lib.cli import identity as osc_id
from osc_lib.cli import parseractions
from osc_lib import exceptions
from osc_lib import utils as osc_utils
from osc_lib.utils import columns as column_util

from openstackclient import command
from openstackclient.i18n import _


LOG = logging.getLogger(__name__)

_attr_map = (
    ('id', 'ID', column_util.LIST_BOTH),
    ('project_id', 'Project', column_util.LIST_LONG_ONLY),
    ('network_id', 'Network ID', column_util.LIST_BOTH),
)
_formatters: dict[str, ty.Any] = {}


def _get_columns(item):
    column_map: dict[str, str] = {}
    hidden_columns = ['location', 'name', 'tenant_id']
    return osc_utils.get_osc_show_columns_for_sdk_resource(
        item, column_map, hidden_columns
    )


class CreateBgpvpnNetAssoc(command.ShowOne):
    _description = _("Create a BGP VPN network association")

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        osc_id.add_project_owner_option_to_parser(parser)
        parser.add_argument(
            'bgpvpn',
            metavar="<bgpvpn>",
            help=_("BGP VPN to apply the network association (name or ID)"),
        )
        parser.add_argument(
            'resource',
            metavar="<network>",
            help=_("Network to associate the BGP VPN (name or ID)"),
        )
        return parser

    def take_action(self, parsed_args):
        client = self.app.client_manager.network
        bgpvpn = client.find_bgpvpn(parsed_args.bgpvpn, ignore_missing=False)
        network = client.find_network(
            parsed_args.resource, ignore_missing=False
        )
        body: dict[str, ty.Any] = {'network_id': network['id']}
        if 'project' in parsed_args and parsed_args.project is not None:
            project_id = osc_id.find_project(
                self.app.client_manager.sdk_connection,
                parsed_args.project,
                parsed_args.project_domain,
            ).id
            body['project_id'] = project_id

        obj = client.create_bgpvpn_network_association(bgpvpn['id'], **body)
        display_columns, columns = _get_columns(obj)
        data = osc_utils.get_dict_properties(
            obj, columns, formatters=_formatters
        )
        return display_columns, data


class DeleteBgpvpnNetAssoc(command.Command):
    _description = _(
        "Delete a BGP VPN network association(s) for a given BGP VPN"
    )

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'resource_association_ids',
            metavar="<network association ID>",
            nargs="+",
            help=_("Network association ID(s) to remove"),
        )
        parser.add_argument(
            'bgpvpn',
            metavar="<bgpvpn>",
            help=_("BGP VPN the network association belongs to (name or ID)"),
        )
        return parser

    def take_action(self, parsed_args):
        client = self.app.client_manager.network
        bgpvpn = client.find_bgpvpn(parsed_args.bgpvpn, ignore_missing=False)
        fails = 0
        for id in parsed_args.resource_association_ids:
            try:
                client.delete_bgpvpn_network_association(bgpvpn['id'], id)
                LOG.warning(
                    "Network association %(id)s deleted",
                    {'id': id},
                )
            except Exception as e:
                fails += 1
                LOG.error(
                    "Failed to delete network "
                    "association with ID '%(id)s': %(e)s",
                    {'id': id, 'e': e},
                )
        if fails > 0:
            msg = _(
                "Failed to delete %(fails)s of %(total)s "
                "network BGP VPN association(s)."
            ) % {
                'fails': fails,
                'total': len(parsed_args.resource_association_ids),
            }
            raise exceptions.CommandError(msg)


class ListBgpvpnNetAssoc(command.Lister):
    _description = _("List BGP VPN network associations for a given BGP VPN")

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'bgpvpn',
            metavar="<bgpvpn>",
            help=_("BGP VPN listed associations belong to (name or ID)"),
        )
        parser.add_argument(
            '--long',
            action='store_true',
            help=_("List additional fields in output"),
        )
        parser.add_argument(
            '--property',
            metavar="<key=value>",
            help=_(
                "Filter property to apply on returned BGP VPNs (repeat to "
                "filter on multiple properties)"
            ),
            action=parseractions.KeyValueAction,
        )
        return parser

    def take_action(self, parsed_args):
        client = self.app.client_manager.network
        bgpvpn = client.find_bgpvpn(parsed_args.bgpvpn, ignore_missing=False)
        params = {}
        if parsed_args.property:
            params.update(parsed_args.property)
        objs = client.bgpvpn_network_associations(
            bgpvpn['id'], retrieve_all=True, **params
        )
        headers, columns = column_util.get_column_definitions(
            list(_attr_map), long_listing=parsed_args.long
        )
        return (
            headers,
            (
                osc_utils.get_dict_properties(
                    s, columns, formatters=_formatters
                )
                for s in objs
            ),
        )


class ShowBgpvpnNetAssoc(command.ShowOne):
    _description = _("Show information of a given BGP VPN network association")

    def get_parser(self, prog_name):
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'resource_association_id',
            metavar="<network association ID>",
            help=_("Network association ID to look up"),
        )
        parser.add_argument(
            'bgpvpn',
            metavar="<bgpvpn>",
            help=_("BGP VPN the association belongs to (name or ID)"),
        )
        return parser

    def take_action(self, parsed_args):
        client = self.app.client_manager.network
        bgpvpn = client.find_bgpvpn(parsed_args.bgpvpn, ignore_missing=False)
        obj = client.get_bgpvpn_network_association(
            bgpvpn['id'], parsed_args.resource_association_id
        )
        display_columns, columns = _get_columns(obj)
        data = osc_utils.get_dict_properties(
            obj, columns, formatters=_formatters
        )
        return display_columns, data



#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

import io
import sys

from openstackclient.common import progressbar
from openstackclient.tests.unit import utils


class TestProgressBarWrapper(utils.TestCase):
    def test_iter_file_display_progress_bar(self):
        size = 98304
        file_obj = io.StringIO('X' * size)
        saved_stdout = sys.stdout
        try:
            sys.stdout = output = FakeTTYStdout()
            file_obj = progressbar.VerboseFileWrapper(file_obj, size)
            chunksize = 1024
            chunk = file_obj.read(chunksize)
            while chunk:
                chunk = file_obj.read(chunksize)
            self.assertEqual('[%s>] 100%%\n' % ('=' * 29), output.getvalue())
        finally:
            sys.stdout = saved_stdout

    def test_iter_file_no_tty(self):
        size = 98304
        file_obj = io.StringIO('X' * size)
        saved_stdout = sys.stdout
        try:
            sys.stdout = output = FakeNoTTYStdout()
            file_obj = progressbar.VerboseFileWrapper(file_obj, size)
            chunksize = 1024
            chunk = file_obj.read(chunksize)
            while chunk:
                chunk = file_obj.read(chunksize)
            # If stdout is not a tty progress bar should do nothing.
            self.assertEqual('', output.getvalue())
        finally:
            sys.stdout = saved_stdout


class FakeTTYStdout(io.StringIO):
    """A Fake stdout that try to emulate a TTY device as much as possible."""

    def isatty(self):
        return True

    def write(self, data):
        # When a CR (carriage return) is found reset file.
        if data.startswith('\r'):
            self.seek(0)
            data = data[1:]
        return io.StringIO.write(self, data)


class FakeNoTTYStdout(FakeTTYStdout):
    """A Fake stdout that is not a TTY device."""

    def isatty(self):
        return False



#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

"""Service Provider action implementations"""

import argparse
from collections.abc import Iterable, Sequence
import logging
from typing import Any

from openstack import utils as sdk_utils
from osc_lib import exceptions
from osc_lib import utils

from openstackclient import command
from openstackclient.i18n import _


LOG = logging.getLogger(__name__)


def _format_service_provider(sp: Any) -> tuple[tuple[str, ...], Any]:
    column_headers = (
        'id',
        'enabled',
        'description',
        'auth_url',
        'sp_url',
        'relay_state_prefix',
    )
    columns = (
        'id',
        'is_enabled',
        'description',
        'auth_url',
        'sp_url',
        'relay_state_prefix',
    )
    return (
        column_headers,
        utils.get_item_properties(sp, columns),
    )


class CreateServiceProvider(command.ShowOne):
    _description = _("Create new service provider")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'service_provider_id',
            metavar='<name>',
            help=_('New service provider name (must be unique)'),
        )
        parser.add_argument(
            '--auth-url',
            metavar='<auth-url>',
            required=True,
            help=_(
                'Authentication URL of remote federated service provider '
                '(required)'
            ),
        )
        parser.add_argument(
            '--description',
            metavar='<description>',
            help=_('New service provider description'),
        )
        parser.add_argument(
            '--service-provider-url',
            metavar='<sp-url>',
            required=True,
            help=_(
                'A service URL where SAML assertions are being sent (required)'
            ),
        )

        enable_service_provider = parser.add_mutually_exclusive_group()
        enable_service_provider.add_argument(
            '--enable',
            dest='is_enabled',
            action='store_true',
            default=True,
            help=_('Enable the service provider (default)'),
        )
        enable_service_provider.add_argument(
            '--disable',
            dest='is_enabled',
            action='store_false',
            help=_('Disable the service provider'),
        )

        return parser

    def take_action(
        self, parsed_args: argparse.Namespace
    ) -> tuple[Sequence[str], Iterable[Any]]:
        service_client = sdk_utils.ensure_service_version(
            self.app.client_manager.sdk_connection.identity, '3'
        )

        kwargs = {}

        kwargs = {'id': parsed_args.service_provider_id}

        if parsed_args.is_enabled is not None:
            kwargs['is_enabled'] = parsed_args.is_enabled

        if parsed_args.description:
            kwargs['description'] = parsed_args.description

        if parsed_args.auth_url:
            kwargs['auth_url'] = parsed_args.auth_url

        if parsed_args.service_provider_url:
            kwargs['sp_url'] = parsed_args.service_provider_url

        sp = service_client.create_service_provider(**kwargs)

        return _format_service_provider(sp)


class DeleteServiceProvider(command.Command):
    _description = _("Delete service provider(s)")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'service_provider',
            metavar='<service-provider>',
            nargs='+',
            help=_('Service provider(s) to delete'),
        )
        return parser

    def take_action(self, parsed_args: argparse.Namespace) -> None:
        service_client = sdk_utils.ensure_service_version(
            self.app.client_manager.sdk_connection.identity, '3'
        )
        result = 0
        for i in parsed_args.service_provider:
            try:
                service_client.delete_service_provider(i)
            except Exception as e:
                result += 1
                LOG.error(
                    _(
                        "Failed to delete service provider with "
                        "name or ID '%(provider)s': %(e)s"
                    ),
                    {'provider': i, 'e': e},
                )

        if result > 0:
            total = len(parsed_args.service_provider)
            msg = _(
                "%(result)s of %(total)s service providers failed to delete."
            ) % {'result': result, 'total': total}
            raise exceptions.CommandError(msg)


class ListServiceProvider(command.Lister):
    _description = _("List service providers")

    def take_action(
        self, parsed_args: argparse.Namespace
    ) -> tuple[tuple[str, ...], Iterable[tuple[Any, ...]]]:
        service_client = sdk_utils.ensure_service_version(
            self.app.client_manager.sdk_connection.identity, '3'
        )
        data = service_client.service_providers()

        column_headers = (
            'ID',
            'Enabled',
            'Description',
            'Auth URL',
            'Service Provider URL',
            'Relay State Prefix',
        )
        columns = (
            'id',
            'is_enabled',
            'description',
            'auth_url',
            'sp_url',
            'relay_state_prefix',
        )
        return (
            column_headers,
            (utils.get_item_properties(s, columns) for s in data),
        )


class SetServiceProvider(command.ShowOne):
    _description = _("Set service provider properties")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'service_provider',
            metavar='<service-provider>',
            help=_('Service provider to modify'),
        )
        parser.add_argument(
            '--auth-url',
            metavar='<auth-url>',
            help=_(
                'New Authentication URL of remote federated service provider'
            ),
        )

        parser.add_argument(
            '--description',
            metavar='<description>',
            help=_('New service provider description'),
        )
        parser.add_argument(
            '--service-provider-url',
            metavar='<sp-url>',
            help=_('New service provider URL, where SAML assertions are sent'),
        )
        enable_service_provider = parser.add_mutually_exclusive_group()
        enable_service_provider.add_argument(
            '--enable',
            dest='is_enabled',
            action='store_true',
            default=None,
            help=_('Enable the service provider'),
        )
        enable_service_provider.add_argument(
            '--disable',
            dest='is_enabled',
            action='store_false',
            default=None,
            help=_('Disable the service provider'),
        )
        return parser

    def take_action(
        self, parsed_args: argparse.Namespace
    ) -> tuple[Sequence[str], Iterable[Any]]:
        service_client = sdk_utils.ensure_service_version(
            self.app.client_manager.sdk_connection.identity, '3'
        )

        kwargs = {}

        if parsed_args.is_enabled is not None:
            kwargs['is_enabled'] = parsed_args.is_enabled

        if parsed_args.description:
            kwargs['description'] = parsed_args.description

        if parsed_args.auth_url:
            kwargs['auth_url'] = parsed_args.auth_url

        if parsed_args.service_provider_url:
            kwargs['sp_url'] = parsed_args.service_provider_url

        service_provider = service_client.update_service_provider(
            parsed_args.service_provider,
            **kwargs,
        )

        return _format_service_provider(service_provider)


class ShowServiceProvider(command.ShowOne):
    _description = _("Display service provider details")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'service_provider',
            metavar='<service-provider>',
            help=_('Service provider to display'),
        )
        return parser

    def take_action(
        self, parsed_args: argparse.Namespace
    ) -> tuple[Sequence[str], Iterable[Any]]:
        service_client = sdk_utils.ensure_service_version(
            self.app.client_manager.sdk_connection.identity, '3'
        )
        service_provider = service_client.find_service_provider(
            parsed_args.service_provider,
            ignore_missing=False,
        )

        return _format_service_provider(service_provider)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from osc_lib.cli import format_columns
from osc_lib import exceptions

from openstackclient.tests.unit.compute.v2 import fakes as compute_fakes
from openstackclient.tests.unit.identity.v3 import fakes as identity_fakes
from openstackclient.tests.unit.volume.v3 import fakes as volume_fakes
from openstackclient.volume.v3 import volume_attachment


class TestVolumeAttachment(volume_fakes.TestVolume):
    def setUp(self):
        super().setUp()

        self.projects_mock = self.identity_client.projects


class TestVolumeAttachmentCreate(TestVolumeAttachment):
    volume = volume_fakes.create_one_volume()
    server = compute_fakes.create_one_server()
    volume_attachment = volume_fakes.create_one_volume_attachment(
        attrs={'instance': server.id, 'volume_id': volume.id},
    )

    columns = (
        'ID',
        'Volume ID',
        'Instance ID',
        'Status',
        'Attach Mode',
        'Attached At',
        'Detached At',
        'Properties',
    )
    data = (
        volume_attachment.id,
        volume_attachment.volume_id,
        volume_attachment.instance,
        volume_attachment.status,
        volume_attachment.attach_mode,
        volume_attachment.attached_at,
        volume_attachment.detached_at,
        format_columns.DictColumn(volume_attachment.connection_info),
    )

    def setUp(self):
        super().setUp()

        self.volume_sdk_client.find_volume.return_value = self.volume
        self.volume_sdk_client.create_attachment.return_value = (
            self.volume_attachment.to_dict()
        )
        self.compute_client.find_server.return_value = self.server

        self.cmd = volume_attachment.CreateVolumeAttachment(self.app, None)

    def test_volume_attachment_create(self):
        self.set_volume_api_version('3.27')

        arglist = [
            self.volume.id,
            self.server.id,
        ]
        verifylist = [
            ('volume', self.volume.id),
            ('server', self.server.id),
            ('connect', False),
            ('initiator', None),
            ('ip', None),
            ('host', None),
            ('platform', None),
            ('os_type', None),
            ('multipath', False),
            ('mountpoint', None),
            ('mode', None),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.volume_sdk_client.find_volume.assert_called_once_with(
            self.volume.id, ignore_missing=False
        )
        self.compute_client.find_server.assert_called_once_with(
            self.server.id, ignore_missing=False
        )
        self.volume_sdk_client.create_attachment.assert_called_once_with(
            self.volume.id,
            connector={},
            instance=self.server.id,
            mode=None,
        )
        self.assertEqual(self.columns, columns)
        self.assertCountEqual(self.data, data)

    def test_volume_attachment_create_with_connect(self):
        self.set_volume_api_version('3.54')

        arglist = [
            self.volume.id,
            self.server.id,
            '--connect',
            '--initiator',
            'iqn.1993-08.org.debian:01:cad181614cec',
            '--ip',
            '192.168.1.20',
            '--host',
            'my-host',
            '--platform',
            'x86_64',
            '--os-type',
            'linux2',
            '--multipath',
            '--mountpoint',
            '/dev/vdb',
            '--mode',
            'null',
        ]
        verifylist = [
            ('volume', self.volume.id),
            ('server', self.server.id),
            ('connect', True),
            ('initiator', 'iqn.1993-08.org.debian:01:cad181614cec'),
            ('ip', '192.168.1.20'),
            ('host', 'my-host'),
            ('platform', 'x86_64'),
            ('os_type', 'linux2'),
            ('multipath', True),
            ('mountpoint', '/dev/vdb'),
            ('mode', 'null'),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        connect_info = dict(
            [
                ('initiator', 'iqn.1993-08.org.debian:01:cad181614cec'),
                ('ip', '192.168.1.20'),
                ('host', 'my-host'),
                ('platform', 'x86_64'),
                ('os_type', 'linux2'),
                ('multipath', True),
                ('mountpoint', '/dev/vdb'),
            ]
        )

        self.volume_sdk_client.find_volume.assert_called_once_with(
            self.volume.id, ignore_missing=False
        )
        self.compute_client.find_server.assert_called_once_with(
            self.server.id, ignore_missing=False
        )
        self.volume_sdk_client.create_attachment.assert_called_once_with(
            self.volume.id,
            connector=connect_info,
            instance=self.server.id,
            mode='null',
        )
        self.assertEqual(self.columns, columns)
        self.assertCountEqual(self.data, data)

    def test_volume_attachment_create_pre_v327(self):
        self.set_volume_api_version('3.26')

        arglist = [
            self.volume.id,
            self.server.id,
        ]
        verifylist = [
            ('volume', self.volume.id),
            ('server', self.server.id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        exc = self.assertRaises(
            exceptions.CommandError, self.cmd.take_action, parsed_args
        )
        self.assertIn(
            '--os-volume-api-version 3.27 or greater is required', str(exc)
        )

    def test_volume_attachment_create_with_mode_pre_v354(self):
        self.set_volume_api_version('3.53')

        arglist = [
            self.volume.id,
            self.server.id,
            '--mode',
            'rw',
        ]
        verifylist = [
            ('volume', self.volume.id),
            ('server', self.server.id),
            ('mode', 'rw'),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        exc = self.assertRaises(
            exceptions.CommandError, self.cmd.take_action, parsed_args
        )
        self.assertIn(
            '--os-volume-api-version 3.54 or greater is required', str(exc)
        )

    def test_volume_attachment_create_with_connect_missing_arg(self):
        self.set_volume_api_version('3.54')

        arglist = [
            self.volume.id,
            self.server.id,
            '--initiator',
            'iqn.1993-08.org.debian:01:cad181614cec',
        ]
        verifylist = [
            ('volume', self.volume.id),
            ('server', self.server.id),
            ('connect', False),
            ('initiator', 'iqn.1993-08.org.debian:01:cad181614cec'),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        exc = self.assertRaises(
            exceptions.CommandError, self.cmd.take_action, parsed_args
        )
        self.assertIn(
            'You must specify the --connect option for any', str(exc)
        )


class TestVolumeAttachmentDelete(TestVolumeAttachment):
    volume_attachment = volume_fakes.create_one_volume_attachment()

    def setUp(self):
        super().setUp()

        self.volume_sdk_client.delete_attachment.return_value = None

        self.cmd = volume_attachment.DeleteVolumeAttachment(self.app, None)

    def test_volume_attachment_delete(self):
        self.set_volume_api_version('3.27')

        arglist = [
            self.volume_attachment.id,
        ]
        verifylist = [
            ('attachment', self.volume_attachment.id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        self.volume_sdk_client.delete_attachment.assert_called_once_with(
            self.volume_attachment.id,
        )
        self.assertIsNone(result)

    def test_volume_attachment_delete_pre_v327(self):
        self.set_volume_api_version('3.26')

        arglist = [
            self.volume_attachment.id,
        ]
        verifylist = [
            ('attachment', self.volume_attachment.id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        exc = self.assertRaises(
            exceptions.CommandError, self.cmd.take_action, parsed_args
        )
        self.assertIn(
            '--os-volume-api-version 3.27 or greater is required', str(exc)
        )


class TestVolumeAttachmentSet(TestVolumeAttachment):
    volume_attachment = volume_fakes.create_one_volume_attachment()

    columns = (
        'ID',
        'Volume ID',
        'Instance ID',
        'Status',
        'Attach Mode',
        'Attached At',
        'Detached At',
        'Properties',
    )
    data = (
        volume_attachment.id,
        volume_attachment.volume_id,
        volume_attachment.instance,
        volume_attachment.status,
        volume_attachment.attach_mode,
        volume_attachment.attached_at,
        volume_attachment.detached_at,
        format_columns.DictColumn(volume_attachment.connection_info),
    )

    def setUp(self):
        super().setUp()

        self.volume_sdk_client.update_attachment.return_value = (
            self.volume_attachment
        )

        self.cmd = volume_attachment.SetVolumeAttachment(self.app, None)

    def test_volume_attachment_set(self):
        self.set_volume_api_version('3.27')

        arglist = [
            self.volume_attachment.id,
            '--initiator',
            'iqn.1993-08.org.debian:01:cad181614cec',
            '--ip',
            '192.168.1.20',
            '--host',
            'my-host',
            '--platform',
            'x86_64',
            '--os-type',
            'linux2',
            '--multipath',
            '--mountpoint',
            '/dev/vdb',
        ]
        verifylist = [
            ('attachment', self.volume_attachment.id),
            ('initiator', 'iqn.1993-08.org.debian:01:cad181614cec'),
            ('ip', '192.168.1.20'),
            ('host', 'my-host'),
            ('platform', 'x86_64'),
            ('os_type', 'linux2'),
            ('multipath', True),
            ('mountpoint', '/dev/vdb'),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        connect_info = dict(
            [
                ('initiator', 'iqn.1993-08.org.debian:01:cad181614cec'),
                ('ip', '192.168.1.20'),
                ('host', 'my-host'),
                ('platform', 'x86_64'),
                ('os_type', 'linux2'),
                ('multipath', True),
                ('mountpoint', '/dev/vdb'),
            ]
        )

        self.volume_sdk_client.update_attachment.assert_called_once_with(
            self.volume_attachment.id,
            connector=connect_info,
        )
        self.assertEqual(self.columns, columns)
        self.assertCountEqual(self.data, data)

    def test_volume_attachment_set_pre_v327(self):
        self.set_volume_api_version('3.26')

        arglist = [
            self.volume_attachment.id,
            '--initiator',
            'iqn.1993-08.org.debian:01:cad181614cec',
        ]
        verifylist = [
            ('attachment', self.volume_attachment.id),
            ('initiator', 'iqn.1993-08.org.debian:01:cad181614cec'),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        exc = self.assertRaises(
            exceptions.CommandError, self.cmd.take_action, parsed_args
        )
        self.assertIn(
            '--os-volume-api-version 3.27 or greater is required', str(exc)
        )


class TestVolumeAttachmentComplete(TestVolumeAttachment):
    volume_attachment = volume_fakes.create_one_volume_attachment()

    def setUp(self):
        super().setUp()

        self.volume_sdk_client.complete_attachment.return_value = None

        self.cmd = volume_attachment.CompleteVolumeAttachment(self.app, None)

    def test_volume_attachment_complete(self):
        self.set_volume_api_version('3.44')

        arglist = [
            self.volume_attachment.id,
        ]
        verifylist = [
            ('attachment', self.volume_attachment.id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        self.volume_sdk_client.complete_attachment.assert_called_once_with(
            self.volume_attachment.id,
        )
        self.assertIsNone(result)

    def test_volume_attachment_complete_pre_v344(self):
        self.set_volume_api_version('3.43')

        arglist = [
            self.volume_attachment.id,
        ]
        verifylist = [
            ('attachment', self.volume_attachment.id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        exc = self.assertRaises(
            exceptions.CommandError, self.cmd.take_action, parsed_args
        )
        self.assertIn(
            '--os-volume-api-version 3.44 or greater is required', str(exc)
        )


class TestVolumeAttachmentList(TestVolumeAttachment):
    project = identity_fakes.FakeProject.create_one_project()
    volume_attachments = volume_fakes.create_volume_attachments()

    columns = (
        'ID',
        'Volume ID',
        'Server ID',
        'Status',
    )
    data = [
        (
            volume_attachment.id,
            volume_attachment.volume_id,
            volume_attachment.instance,
            volume_attachment.status,
        )
        for volume_attachment in volume_attachments
    ]

    def setUp(self):
        super().setUp()

        self.projects_mock.get.return_value = self.project
        self.volume_sdk_client.attachments.return_value = (
            self.volume_attachments
        )

        self.cmd = volume_attachment.ListVolumeAttachment(self.app, None)

    def test_volume_attachment_list(self):
        self.set_volume_api_version('3.27')

        arglist = []
        verifylist = [
            ('project', None),
            ('all_projects', False),
            ('volume_id', None),
            ('status', None),
            ('marker', None),
            ('limit', None),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.volume_sdk_client.attachments.assert_called_once_with(
            search_opts={
                'all_tenants': False,
                'project_id': None,
                'status': None,
                'volume_id': None,
            },
            marker=None,
            limit=None,
        )
        self.assertEqual(self.columns, columns)
        self.assertCountEqual(tuple(self.data), data)

    def test_volume_attachment_list_with_options(self):
        self.set_volume_api_version('3.27')

        arglist = [
            '--project',
            self.project.name,
            '--volume-id',
            'volume-id',
            '--status',
            'attached',
            '--marker',
            'volume-attachment-id',
            '--limit',
            '2',
        ]
        verifylist = [
            ('project', self.project.name),
            ('all_projects', False),
            ('volume_id', 'volume-id'),
            ('status', 'attached'),
            ('marker', 'volume-attachment-id'),
            ('limit', 2),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.volume_sdk_client.attachments.assert_called_once_with(
            search_opts={
                'all_tenants': True,
                'project_id': self.project.id,
                'status': 'attached',
                'volume_id': 'volume-id',
            },
            marker='volume-attachment-id',
            limit=2,
        )
        self.assertEqual(self.columns, columns)
        self.assertCountEqual(tuple(self.data), data)

    def test_volume_attachment_list_pre_v327(self):
        self.set_volume_api_version('3.26')

        arglist = []
        verifylist = [
            ('project', None),
            ('all_projects', False),
            ('volume_id', None),
            ('status', None),
            ('marker', None),
            ('limit', None),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        exc = self.assertRaises(
            exceptions.CommandError, self.cmd.take_action, parsed_args
        )
        self.assertIn(
            '--os-volume-api-version 3.27 or greater is required', str(exc)
        )



#   Copyright 2013 Nebula Inc.
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

"""Test module module"""

import sys
from unittest import mock

from openstackclient.common import module as osc_module
from openstackclient.tests.unit import utils


class FakeModule:
    def __init__(self, name, version):
        self.name = name
        self.__version__ = version
        # Workaround for openstacksdk case
        self.version = mock.Mock()
        self.version.__version__ = version


# NOTE(dtroyer): module_1 must match the version list filter (not --all)
#                currently == '*client*'
module_name_1 = 'fakeclient'
module_version_1 = '0.1.2'

module_name_2 = 'zlib'
module_version_2 = '1.1'

# module_3 match openstacksdk
module_name_3 = 'openstack'
module_version_3 = '0.9.13'

# module_4 match sub module of fakeclient
module_name_4 = 'fakeclient.submodule'
module_version_4 = '0.2.2'

# module_5 match private module
module_name_5 = '_private_module.lib'
module_version_5 = '0.0.1'

MODULES = {
    'sys': sys,
    module_name_1: FakeModule(module_name_1, module_version_1),
    module_name_2: FakeModule(module_name_2, module_version_2),
    module_name_3: FakeModule(module_name_3, module_version_3),
    module_name_4: FakeModule(module_name_4, module_version_4),
    module_name_5: FakeModule(module_name_5, module_version_5),
}


class TestCommandList(utils.TestCommand):
    def setUp(self):
        super().setUp()

        self.app.command_manager = mock.Mock()
        self.app.command_manager.get_command_groups.return_value = [
            'openstack.common'
        ]
        self.app.command_manager.get_command_names.return_value = [
            'limits show\nextension list'
        ]

        # Get the command object to test
        self.cmd = osc_module.ListCommand(self.app, None)

    def test_command_list_no_options(self):
        arglist = []
        verifylist = []
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class Lister in cliff, abstract method take_action()
        # returns a tuple containing the column names and an iterable
        # containing the data to be listed.
        columns, data = self.cmd.take_action(parsed_args)

        # TODO(bapalm): Adjust this when cliff properly supports
        # handling the detection rather than using the hard-code below.
        collist = ('Command Group', 'Commands')
        self.assertEqual(collist, columns)
        datalist = (('openstack.common', 'limits show\nextension list'),)

        self.assertEqual(datalist, tuple(data))

    def test_command_list_with_group_not_found(self):
        arglist = [
            '--group',
            'not_exist',
        ]
        verifylist = [
            ('group', 'not_exist'),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        collist = ('Command Group', 'Commands')
        self.assertEqual(collist, columns)
        self.assertEqual([], data)

    def test_command_list_with_group(self):
        arglist = [
            '--group',
            'common',
        ]
        verifylist = [
            ('group', 'common'),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        collist = ('Command Group', 'Commands')
        self.assertEqual(collist, columns)
        datalist = (('openstack.common', 'limits show\nextension list'),)

        self.assertEqual(datalist, tuple(data))


class TestModuleList(utils.TestCommand):
    def setUp(self):
        super().setUp()

        # Get the command object to test
        self.cmd = osc_module.ListModule(self.app, None)

    def test_module_list_no_options(self):
        arglist = []
        verifylist = [
            ('all', False),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class Lister in cliff, abstract method take_action()
        # returns a tuple containing the column names and an iterable
        # containing the data to be listed.
        with mock.patch.dict(
            'openstackclient.common.module.sys.modules',
            values=MODULES,
            clear=True,
        ):
            columns, data = self.cmd.take_action(parsed_args)

        # Output xxxclient and openstacksdk, but not regular module, like: zlib
        self.assertIn(module_name_1, columns)
        self.assertIn(module_version_1, data)
        self.assertNotIn(module_name_2, columns)
        self.assertNotIn(module_version_2, data)
        self.assertIn(module_name_3, columns)
        self.assertIn(module_version_3, data)
        # Filter sub and private modules
        self.assertNotIn(module_name_4, columns)
        self.assertNotIn(module_version_4, data)
        self.assertNotIn(module_name_5, columns)
        self.assertNotIn(module_version_5, data)

    def test_module_list_all(self):
        arglist = [
            '--all',
        ]
        verifylist = [
            ('all', True),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class Lister in cliff, abstract method take_action()
        # returns a tuple containing the column names and an iterable
        # containing the data to be listed.
        with mock.patch.dict(
            'openstackclient.common.module.sys.modules',
            values=MODULES,
            clear=True,
        ):
            columns, data = self.cmd.take_action(parsed_args)

        # Output xxxclient, openstacksdk and regular module, like: zlib
        self.assertIn(module_name_1, columns)
        self.assertIn(module_version_1, data)
        self.assertIn(module_name_2, columns)
        self.assertIn(module_version_2, data)
        self.assertIn(module_name_3, columns)
        self.assertIn(module_version_3, data)
        # Filter sub and private modules
        self.assertNotIn(module_name_4, columns)
        self.assertNotIn(module_version_4, data)
        self.assertNotIn(module_name_5, columns)
        self.assertNotIn(module_version_5, data)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from tempest.lib.common.utils import data_utils

from openstackclient.tests.functional.identity.v2 import common


class ProjectTests(common.IdentityTests):
    def test_project_create(self):
        project_name = data_utils.rand_name('TestProject')
        description = data_utils.rand_name('description')
        raw_output = self.openstack(
            'project create '
            f'--description {description} '
            '--enable '
            '--property k1=v1 '
            '--property k2=v2 '
            f'{project_name}'
        )
        self.addCleanup(self.openstack, f'project delete {project_name}')
        items = self.parse_show(raw_output)
        show_fields = list(self.PROJECT_FIELDS)
        show_fields.extend(['k1', 'k2'])
        self.assert_show_fields(items, show_fields)
        project = self.parse_show_as_object(raw_output)
        self.assertEqual('v1', project['k1'])
        self.assertEqual('v2', project['k2'])

    def test_project_delete(self):
        project_name = self._create_dummy_project(add_clean_up=False)
        raw_output = self.openstack(f'project delete {project_name}')
        self.assertEqual(0, len(raw_output))

    def test_project_list(self):
        raw_output = self.openstack('project list')
        items = self.parse_listing(raw_output)
        self.assert_table_structure(items, common.BASIC_LIST_HEADERS)

    def test_project_set(self):
        project_name = self._create_dummy_project()
        new_project_name = data_utils.rand_name('NewTestProject')
        raw_output = self.openstack(
            'project set '
            f'--name {new_project_name} '
            '--disable '
            '--property k0=v0 '
            f'{project_name}'
        )
        self.assertEqual(0, len(raw_output))
        # check project details
        raw_output = self.openstack(f'project show {new_project_name}')
        items = self.parse_show(raw_output)
        fields = list(self.PROJECT_FIELDS)
        fields.extend(['properties'])
        self.assert_show_fields(items, fields)
        project = self.parse_show_as_object(raw_output)
        self.assertEqual(new_project_name, project['name'])
        self.assertFalse(project['enabled'])
        self.assertEqual("k0='v0'", project['properties'])

    def test_project_show(self):
        project_name = self._create_dummy_project()
        raw_output = self.openstack(f'project show {project_name}')
        items = self.parse_show(raw_output)
        fields = list(self.PROJECT_FIELDS)
        fields.extend(['properties'])
        self.assert_show_fields(items, fields)



#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

import uuid

from osc_lib import exceptions

from openstackclient.compute.v2 import server_share
from openstackclient.tests.unit.compute.v2 import fakes as compute_fakes


class TestServerShareList(compute_fakes.TestComputev2):
    def setUp(self):
        super().setUp()

        self.set_compute_api_version('2.97')

        self.server = compute_fakes.create_one_server()
        self.shares = compute_fakes.create_shares()

        self.compute_client.find_server.return_value = self.server
        self.compute_client.share_attachments.return_value = self.shares

        self.cmd = server_share.ListServerShare(self.app, None)

    def test_server_share_list(self):
        arglist = [
            self.server.id,
        ]
        verifylist = [
            ('server', self.server.id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)

        self.assertEqual(("Share ID", "Status", "Tag"), columns)
        self.assertEqual(
            (
                (
                    self.shares[0].share_id,
                    self.shares[0].status,
                    self.shares[0].tag,
                ),
                (
                    self.shares[1].share_id,
                    self.shares[1].status,
                    self.shares[1].tag,
                ),
            ),
            tuple(data),
        )
        self.compute_client.share_attachments.assert_called_once_with(
            self.server,
        )

    def test_server_share_list_pre_v297(self):
        self.set_compute_api_version('2.96')
        arglist = [self.server.id]
        verifylist = [('server', self.server.id)]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        self.assertRaises(
            exceptions.CommandError,
            self.cmd.take_action,
            parsed_args,
        )


class TestServerShareShow(compute_fakes.TestComputev2):
    def setUp(self):
        super().setUp()

        self.set_compute_api_version('2.97')

        self.server = compute_fakes.create_one_server()
        self.share = compute_fakes.create_one_share()

        self.compute_client.find_server.return_value = self.server
        self.compute_client.get_share_attachment.return_value = self.share

        self.shared_file_system_client = (
            self.app.client_manager.sdk_connection.shared_file_system
        )
        self._manila_share_id = uuid.uuid4().hex
        self.shared_file_system_client.find_share.return_value = type(
            'FakeShare', (), {'id': self._manila_share_id}
        )()

        self.cmd = server_share.ShowServerShare(self.app, None)

    def test_server_share_show(self):
        arglist = [
            self.server.id,
            self._manila_share_id,
        ]
        verifylist = [
            ('server', self.server.id),
            ('share', self._manila_share_id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)

        self.assertEqual(
            ('Export Location', 'Share ID', 'Status', 'Tag', 'UUID'), columns
        )
        self.assertEqual(
            (
                self.share.export_location,
                self.share.share_id,
                self.share.status,
                self.share.tag,
                self.share.uuid,
            ),
            tuple(data),
        )
        self.shared_file_system_client.find_share.assert_called_once_with(
            self._manila_share_id, ignore_missing=False
        )
        self.compute_client.get_share_attachment.assert_called_once_with(
            self.server, self._manila_share_id
        )

    def test_server_share_show_pre_v297(self):
        self.set_compute_api_version('2.96')
        arglist = [self.server.id, self._manila_share_id]
        verifylist = [
            ('server', self.server.id),
            ('share', self._manila_share_id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        self.assertRaises(
            exceptions.CommandError,
            self.cmd.take_action,
            parsed_args,
        )


class TestServerShareCreate(compute_fakes.TestComputev2):
    def setUp(self):
        super().setUp()

        self.set_compute_api_version('2.97')

        self.server = compute_fakes.create_one_server()
        self.share = compute_fakes.create_one_share()

        self.compute_client.find_server.return_value = self.server
        self.compute_client.create_share_attachment.return_value = self.share

        self.shared_file_system_client = (
            self.app.client_manager.sdk_connection.shared_file_system
        )
        self._manila_share_id = uuid.uuid4().hex
        self.shared_file_system_client.find_share.return_value = type(
            'FakeShare', (), {'id': self._manila_share_id}
        )()

        self.cmd = server_share.AddServerShare(self.app, None)

    def test_server_share_create(self):
        arglist = [
            self.server.id,
            self._manila_share_id,
            "--tag",
            "my-tag",
        ]
        verifylist = [
            ('server', self.server.id),
            ('share', self._manila_share_id),
            ('tag', 'my-tag'),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)

        self.assertEqual(
            ("Export Location", "Share ID", "Status", "Tag", "UUID"), columns
        )
        self.assertEqual(
            (
                self.share.export_location,
                self.share.share_id,
                self.share.status,
                self.share.tag,
                self.share.uuid,
            ),
            tuple(data),
        )
        self.shared_file_system_client.find_share.assert_called_once_with(
            self._manila_share_id, ignore_missing=False
        )
        self.compute_client.create_share_attachment.assert_called_once_with(
            self.server, self._manila_share_id, tag='my-tag'
        )

    def test_server_share_create_no_tag(self):
        arglist = [
            self.server.id,
            self._manila_share_id,
        ]
        verifylist = [
            ('server', self.server.id),
            ('share', self._manila_share_id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        self.cmd.take_action(parsed_args)

        self.compute_client.create_share_attachment.assert_called_once_with(
            self.server, self._manila_share_id
        )

    def test_server_share_create_pre_v297(self):
        self.set_compute_api_version('2.96')
        arglist = [self.server.id, self._manila_share_id]
        verifylist = [
            ('server', self.server.id),
            ('share', self._manila_share_id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        self.assertRaises(
            exceptions.CommandError,
            self.cmd.take_action,
            parsed_args,
        )


class TestServerShareDelete(compute_fakes.TestComputev2):
    def setUp(self):
        super().setUp()

        self.set_compute_api_version('2.97')

        self.server = compute_fakes.create_one_server()

        self.compute_client.find_server.return_value = self.server

        self.shared_file_system_client = (
            self.app.client_manager.sdk_connection.shared_file_system
        )
        self._manila_share_id = uuid.uuid4().hex
        self.shared_file_system_client.find_share.return_value = type(
            'FakeShare', (), {'id': self._manila_share_id}
        )()

        self.cmd = server_share.RemoveServerShare(self.app, None)

    def test_server_share_delete(self):
        arglist = [
            self.server.id,
            self._manila_share_id,
        ]
        verifylist = [
            ('server', self.server.id),
            ('share', self._manila_share_id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        self.cmd.take_action(parsed_args)

        self.shared_file_system_client.find_share.assert_called_once_with(
            self._manila_share_id, ignore_missing=False
        )
        self.compute_client.delete_share_attachment.assert_called_once_with(
            self.server,
            self._manila_share_id,
        )

    def test_server_share_delete_pre_v297(self):
        self.set_compute_api_version('2.96')
        arglist = [self.server.id, self._manila_share_id]
        verifylist = [
            ('server', self.server.id),
            ('share', self._manila_share_id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        self.assertRaises(
            exceptions.CommandError,
            self.cmd.take_action,
            parsed_args,
        )



#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

"""Object Store v1 API Library Tests"""

from unittest import mock

from keystoneauth1 import session
from requests_mock.contrib import fixture

from openstackclient.api import object_store_v1 as object_store
from openstackclient.tests.unit import utils


FAKE_ACCOUNT = 'q12we34r'
FAKE_AUTH = '11223344556677889900'
FAKE_URL = 'http://gopher.com/v1/' + FAKE_ACCOUNT

FAKE_CONTAINER = 'rainbarrel'
FAKE_OBJECT = 'spigot'

LIST_CONTAINER_RESP = [
    {
        "name": "qaz",
        "count": 0,
        "bytes": 0,
        "last_modified": "2020-05-16T05:52:07.377550",
    },
    {
        "name": "fred",
        "count": 0,
        "bytes": 0,
        "last_modified": "2020-05-16T05:55:07.377550",
    },
]

LIST_OBJECT_RESP = [
    {'name': 'fred', 'bytes': 1234, 'content_type': 'text'},
    {'name': 'wilma', 'bytes': 5678, 'content_type': 'text'},
]


class TestObjectAPIv1(utils.TestCase):
    def setUp(self):
        super().setUp()
        sess = session.Session()
        self.api = object_store.APIv1(session=sess, endpoint=FAKE_URL)
        self.requests_mock = self.useFixture(fixture.Fixture())


class TestContainer(TestObjectAPIv1):
    def setUp(self):
        super().setUp()

    def test_container_create(self):
        headers = {
            'x-trans-id': '1qaz2wsx',
        }
        self.requests_mock.register_uri(
            'PUT',
            FAKE_URL + '/qaz',
            headers=headers,
            status_code=201,
        )
        ret = self.api.container_create(container='qaz')
        data = {
            'account': FAKE_ACCOUNT,
            'container': 'qaz',
            'x-trans-id': '1qaz2wsx',
        }
        self.assertEqual(data, ret)

    def test_container_delete(self):
        self.requests_mock.register_uri(
            'DELETE',
            FAKE_URL + '/qaz',
            status_code=204,
        )
        ret = self.api.container_delete(container='qaz')
        self.assertIsNone(ret)

    def test_container_list_no_options(self):
        self.requests_mock.register_uri(
            'GET',
            FAKE_URL,
            json=LIST_CONTAINER_RESP,
            status_code=200,
        )
        ret = self.api.container_list()
        self.assertEqual(LIST_CONTAINER_RESP, ret)

    def test_container_list_prefix(self):
        self.requests_mock.register_uri(
            'GET',
            FAKE_URL + '?prefix=foo%2f&format=json',
            json=LIST_CONTAINER_RESP,
            status_code=200,
        )
        ret = self.api.container_list(
            prefix='foo/',
        )
        self.assertEqual(LIST_CONTAINER_RESP, ret)

    def test_container_list_marker_limit_end(self):
        self.requests_mock.register_uri(
            'GET',
            FAKE_URL + '?marker=next&limit=2&end_marker=stop&format=json',
            json=LIST_CONTAINER_RESP,
            status_code=200,
        )
        ret = self.api.container_list(
            marker='next',
            limit=2,
            end_marker='stop',
        )
        self.assertEqual(LIST_CONTAINER_RESP, ret)

    def test_container_list_full_listing(self):
        self.requests_mock.register_uri(
            'GET',
            FAKE_URL + '?limit=1&format=json',
            json=[LIST_CONTAINER_RESP[0]],
            status_code=200,
        )
        self.requests_mock.register_uri(
            'GET',
            FAKE_URL
            + '?marker={}&limit=1&format=json'.format(
                LIST_CONTAINER_RESP[0]['name']
            ),
            json=[LIST_CONTAINER_RESP[1]],
            status_code=200,
        )
        self.requests_mock.register_uri(
            'GET',
            FAKE_URL
            + '?marker={}&limit=1&format=json'.format(
                LIST_CONTAINER_RESP[1]['name']
            ),
            json=[],
            status_code=200,
        )
        ret = self.api.container_list(
            limit=1,
            full_listing=True,
        )
        self.assertEqual(LIST_CONTAINER_RESP, ret)

    def test_container_show(self):
        headers = {
            'X-Container-Meta-Owner': FAKE_ACCOUNT,
            'x-container-object-count': '1',
            'x-container-bytes-used': '577',
            'x-storage-policy': 'o1--sr-r3',
        }
        resp = {
            'account': FAKE_ACCOUNT,
            'container': 'qaz',
            'object_count': '1',
            'bytes_used': '577',
            'storage_policy': 'o1--sr-r3',
            'properties': {'Owner': FAKE_ACCOUNT},
        }
        self.requests_mock.register_uri(
            'HEAD',
            FAKE_URL + '/qaz',
            headers=headers,
            status_code=204,
        )
        ret = self.api.container_show(container='qaz')
        self.assertEqual(resp, ret)


class TestObject(TestObjectAPIv1):
    def setUp(self):
        super().setUp()

    @mock.patch('openstackclient.api.object_store_v1.open')
    def base_object_create(self, file_contents, mock_open):
        mock_open.read.return_value = file_contents

        headers = {
            'etag': 'youreit',
            'x-trans-id': '1qaz2wsx',
        }
        # TODO(dtroyer): When requests_mock gains the ability to
        #                match against request.body add this check
        #                https://review.opendev.org/127316
        self.requests_mock.register_uri(
            'PUT',
            FAKE_URL + '/qaz/counter.txt',
            headers=headers,
            # body=file_contents,
            status_code=201,
        )
        ret = self.api.object_create(
            container='qaz',
            object='counter.txt',
        )
        data = {
            'account': FAKE_ACCOUNT,
            'container': 'qaz',
            'object': 'counter.txt',
            'etag': 'youreit',
            'x-trans-id': '1qaz2wsx',
        }
        self.assertEqual(data, ret)

    def test_object_create(self):
        self.base_object_create('111\n222\n333\n')
        self.base_object_create(bytes([0x31, 0x00, 0x0D, 0x0A, 0x7F, 0xFF]))

    def test_object_delete(self):
        self.requests_mock.register_uri(
            'DELETE',
            FAKE_URL + '/qaz/wsx',
            status_code=204,
        )
        ret = self.api.object_delete(
            container='qaz',
            object='wsx',
        )
        self.assertIsNone(ret)

    def test_object_list_no_options(self):
        self.requests_mock.register_uri(
            'GET',
            FAKE_URL + '/qaz',
            json=LIST_OBJECT_RESP,
            status_code=200,
        )
        ret = self.api.object_list(container='qaz')
        self.assertEqual(LIST_OBJECT_RESP, ret)

    def test_object_list_delimiter(self):
        self.requests_mock.register_uri(
            'GET',
            FAKE_URL + '/qaz?delimiter=%7C',
            json=LIST_OBJECT_RESP,
            status_code=200,
        )
        ret = self.api.object_list(
            container='qaz',
            delimiter='|',
        )
        self.assertEqual(LIST_OBJECT_RESP, ret)

    def test_object_list_prefix(self):
        self.requests_mock.register_uri(
            'GET',
            FAKE_URL + '/qaz?prefix=foo%2f',
            json=LIST_OBJECT_RESP,
            status_code=200,
        )
        ret = self.api.object_list(
            container='qaz',
            prefix='foo/',
        )
        self.assertEqual(LIST_OBJECT_RESP, ret)

    def test_object_list_marker_limit_end(self):
        self.requests_mock.register_uri(
            'GET',
            FAKE_URL + '/qaz?marker=next&limit=2&end_marker=stop',
            json=LIST_CONTAINER_RESP,
            status_code=200,
        )
        ret = self.api.object_list(
            container='qaz',
            marker='next',
            limit=2,
            end_marker='stop',
        )
        self.assertEqual(LIST_CONTAINER_RESP, ret)

    #     def test_list_objects_full_listing(self):
    #         sess = self.app.client_manager.session
    #
    #         def side_effect(*args, **kwargs):
    #             rv = sess.get().json.return_value
    #             sess.get().json.return_value = []
    #             sess.get().json.side_effect = None
    #             return rv
    #
    #         resp = [{'name': 'is-name'}]
    #         sess.get().json.return_value = resp
    #         sess.get().json.side_effect = side_effect
    #
    #         data = lib_object.list_objects(
    #             sess,
    #             fake_url,
    #             fake_container,
    #             full_listing=True,
    #         )
    #
    #         # Check expected values
    #         sess.get.assert_called_with(
    #             fake_url + '/' + fake_container,
    #             params={
    #                 'format': 'json',
    #                 'marker': 'is-name',
    #             }
    #         )
    #         self.assertEqual(resp, data)

    def test_object_show(self):
        headers = {
            'content-type': 'text/alpha',
            'content-length': '577',
            'last-modified': '20130101',
            'etag': 'qaz',
            'x-container-meta-owner': FAKE_ACCOUNT,
            'x-object-meta-wife': 'Wilma',
            'x-object-meta-Husband': 'fred',
            'x-tra-header': 'yabba-dabba-do',
        }
        resp = {
            'account': FAKE_ACCOUNT,
            'container': 'qaz',
            'object': FAKE_OBJECT,
            'content-type': 'text/alpha',
            'content-length': '577',
            'last-modified': '20130101',
            'etag': 'qaz',
            'properties': {'wife': 'Wilma', 'Husband': 'fred'},
        }
        self.requests_mock.register_uri(
            'HEAD',
            FAKE_URL + '/qaz/' + FAKE_OBJECT,
            headers=headers,
            status_code=204,
        )
        ret = self.api.object_show(
            container='qaz',
            object=FAKE_OBJECT,
        )
        self.assertEqual(resp, ret)



#   Copyright 2013 Nebula Inc.
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.

from unittest import mock
import uuid

from openstack.image.v1 import _proxy
from openstack.image.v1 import image

from openstackclient.tests.unit import fakes
from openstackclient.tests.unit import utils
from openstackclient.tests.unit.volume.v2 import fakes as volume_fakes


class FakeClientMixin:
    def setUp(self):
        super().setUp()

        self.app.client_manager.image = mock.Mock(spec=_proxy.Proxy)
        self.image_client = self.app.client_manager.image


class TestImagev1(FakeClientMixin, utils.TestCommand):
    def setUp(self):
        super().setUp()

        self.app.client_manager.volume = volume_fakes.FakeVolumeClient(
            endpoint=fakes.AUTH_URL,
            token=fakes.AUTH_TOKEN,
        )
        self.volume_client = self.app.client_manager.volume


def create_one_image(attrs=None):
    """Create a fake image.

    :param Dictionary attrs:
        A dictionary with all attributes of image
    :return:
        A FakeResource object with id, name, owner, protected,
        visibility and tags attrs
    """
    attrs = attrs or {}

    # Set default attribute
    image_info = {
        'id': str(uuid.uuid4()),
        'name': 'image-name' + uuid.uuid4().hex,
        'owner': 'image-owner' + uuid.uuid4().hex,
        'container_format': '',
        'disk_format': '',
        'min_disk': 0,
        'min_ram': 0,
        'is_public': True,
        'protected': False,
        'properties': {'Alpha': 'a', 'Beta': 'b', 'Gamma': 'g'},
        'status': 'status' + uuid.uuid4().hex,
    }

    # Overwrite default attributes if there are some attributes set
    image_info.update(attrs)

    return image.Image(**image_info)



#   Copyright 2014 CERN.
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.


from openstack.identity.v3 import service_provider as _service_provider
from openstack.test import fakes as sdk_fakes

from openstackclient.identity.v3 import service_provider
from openstackclient.tests.unit.identity.v3 import fakes as service_fakes


class TestServiceProviderCreate(service_fakes.TestFederatedIdentity):
    columns = (
        'id',
        'enabled',
        'description',
        'auth_url',
        'sp_url',
        'relay_state_prefix',
    )

    def setUp(self):
        super().setUp()

        self.service_provider = sdk_fakes.generate_fake_resource(
            _service_provider.ServiceProvider
        )
        self.identity_sdk_client.create_service_provider.return_value = (
            self.service_provider
        )
        self.data = (
            self.service_provider.id,
            self.service_provider.is_enabled,
            self.service_provider.description,
            self.service_provider.auth_url,
            self.service_provider.sp_url,
            self.service_provider.relay_state_prefix,
        )
        self.cmd = service_provider.CreateServiceProvider(self.app, None)

    def test_create_service_provider_required_options_only(self):
        arglist = [
            '--auth-url',
            self.service_provider.auth_url,
            '--service-provider-url',
            self.service_provider.sp_url,
            self.service_provider.id,
        ]
        verifylist = [
            ('auth_url', self.service_provider.auth_url),
            ('service_provider_url', self.service_provider.sp_url),
            ('service_provider_id', self.service_provider.id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)

        # Set expected values
        kwargs = {
            'is_enabled': True,
            'auth_url': self.service_provider.auth_url,
            'sp_url': self.service_provider.sp_url,
        }

        self.identity_sdk_client.create_service_provider.assert_called_with(
            id=self.service_provider.id, **kwargs
        )

        self.assertEqual(self.columns, columns)
        self.assertEqual(self.data, data)

    def test_create_service_provider_description(self):
        arglist = [
            '--description',
            self.service_provider.description,
            '--auth-url',
            self.service_provider.auth_url,
            '--service-provider-url',
            self.service_provider.sp_url,
            self.service_provider.id,
        ]
        verifylist = [
            ('description', self.service_provider.description),
            ('auth_url', self.service_provider.auth_url),
            ('service_provider_url', self.service_provider.sp_url),
            ('service_provider_id', self.service_provider.id),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)

        # Set expected values
        kwargs = {
            'description': self.service_provider.description,
            'auth_url': self.service_provider.auth_url,
            'sp_url': self.service_provider.sp_url,
            'is_enabled': self.service_provider.is_enabled,
        }

        self.identity_sdk_client.create_service_provider.assert_called_with(
            id=self.service_provider.id, **kwargs
        )

        self.assertEqual(self.columns, columns)
        self.assertEqual(self.data, data)

    def test_create_service_provider_disabled(self):
        arglist = [
            '--auth-url',
            self.service_provider.auth_url,
            '--service-provider-url',
            self.service_provider.sp_url,
            '--disable',
            self.service_provider.id,
        ]
        verifylist = [
            ('auth_url', self.service_provider.auth_url),
            ('service_provider_url', self.service_provider.sp_url),
            ('service_provider_id', self.service_provider.id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)
        # Set expected values
        kwargs = {
            'auth_url': self.service_provider.auth_url,
            'sp_url': self.service_provider.sp_url,
            'is_enabled': False,
        }

        self.identity_sdk_client.create_service_provider.assert_called_with(
            id=self.service_provider.id, **kwargs
        )
        self.assertEqual(self.columns, columns)
        self.assertEqual(self.data, data)


class TestServiceProviderDelete(service_fakes.TestFederatedIdentity):
    def setUp(self):
        super().setUp()

        self.service_provider = sdk_fakes.generate_fake_resource(
            _service_provider.ServiceProvider
        )
        self.identity_sdk_client.delete_service_provider.return_value = None
        self.cmd = service_provider.DeleteServiceProvider(self.app, None)

    def test_delete_service_provider(self):
        arglist = [
            self.service_provider.id,
        ]
        verifylist = [
            ('service_provider', [self.service_provider.id]),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        self.identity_sdk_client.delete_service_provider.assert_called_with(
            self.service_provider.id,
        )
        self.assertIsNone(result)


class TestServiceProviderList(service_fakes.TestFederatedIdentity):
    def setUp(self):
        super().setUp()

        self.service_provider = sdk_fakes.generate_fake_resource(
            _service_provider.ServiceProvider
        )
        self.identity_sdk_client.service_providers.return_value = [
            self.service_provider
        ]

        # Get the command object to test
        self.cmd = service_provider.ListServiceProvider(self.app, None)

    def test_service_provider_list_no_options(self):
        arglist = []
        verifylist = []
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        # In base command class Lister in cliff, abstract method take_action()
        # returns a tuple containing the column names and an iterable
        # containing the data to be listed.
        columns, data = self.cmd.take_action(parsed_args)

        self.identity_sdk_client.service_providers.assert_called_with()

        collist = (
            'ID',
            'Enabled',
            'Description',
            'Auth URL',
            'Service Provider URL',
            'Relay State Prefix',
        )
        self.assertEqual(collist, columns)
        datalist = (
            (
                self.service_provider.id,
                True,
                self.service_provider.description,
                self.service_provider.auth_url,
                self.service_provider.sp_url,
                self.service_provider.relay_state_prefix,
            ),
        )
        self.assertEqual(tuple(data), datalist)


class TestServiceProviderSet(service_fakes.TestFederatedIdentity):
    columns = (
        'id',
        'enabled',
        'description',
        'auth_url',
        'sp_url',
        'relay_state_prefix',
    )

    def setUp(self):
        super().setUp()
        self.service_provider = sdk_fakes.generate_fake_resource(
            _service_provider.ServiceProvider
        )
        self.identity_sdk_client.update_service_provider.return_value = (
            self.service_provider
        )
        self.data = (
            self.service_provider.id,
            self.service_provider.is_enabled,
            self.service_provider.description,
            self.service_provider.auth_url,
            self.service_provider.sp_url,
            self.service_provider.relay_state_prefix,
        )
        self.cmd = service_provider.SetServiceProvider(self.app, None)

    def test_service_provider_disable(self):
        """Disable Service Provider

        Set Service Provider's ``enabled`` attribute to False.
        """
        arglist = [
            '--disable',
            self.service_provider.id,
        ]
        verifylist = [
            ('service_provider', self.service_provider.id),
            ('is_enabled', False),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)
        self.identity_sdk_client.update_service_provider.assert_called_with(
            self.service_provider.id,
            is_enabled=False,
        )
        self.assertEqual(columns, self.columns)
        self.assertEqual(data, self.data)

    def test_service_provider_enable(self):
        """Enable Service Provider.

        Set Service Provider's ``enabled`` attribute to True.
        """
        arglist = [
            '--enable',
            self.service_provider.id,
        ]
        verifylist = [
            ('service_provider', self.service_provider.id),
            ('is_enabled', True),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)
        self.identity_sdk_client.update_service_provider.assert_called_with(
            self.service_provider.id,
            is_enabled=True,
        )
        self.assertEqual(columns, self.columns)
        self.assertEqual(data, self.data)

    def test_service_provider_no_options(self):
        arglist = [
            self.service_provider.id,
        ]
        verifylist = [
            ('service_provider', self.service_provider.id),
            ('description', None),
            ('is_enabled', None),
            ('auth_url', None),
            ('service_provider_url', None),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)
        self.assertEqual(columns, self.columns)
        self.assertEqual(data, self.data)


class TestServiceProviderShow(service_fakes.TestFederatedIdentity):
    def setUp(self):
        super().setUp()

        self.service_provider = sdk_fakes.generate_fake_resource(
            _service_provider.ServiceProvider
        )
        self.identity_sdk_client.find_service_provider.return_value = (
            self.service_provider
        )
        self.data = (
            self.service_provider.id,
            self.service_provider.is_enabled,
            self.service_provider.description,
            self.service_provider.auth_url,
            self.service_provider.sp_url,
            self.service_provider.relay_state_prefix,
        )

        # Get the command object to test
        self.cmd = service_provider.ShowServiceProvider(self.app, None)

    def test_service_provider_show(self):
        arglist = [
            self.service_provider.id,
        ]
        verifylist = [
            ('service_provider', self.service_provider.id),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.identity_sdk_client.find_service_provider.assert_called_with(
            self.service_provider.id,
            ignore_missing=False,
        )

        collist = (
            'id',
            'enabled',
            'description',
            'auth_url',
            'sp_url',
            'relay_state_prefix',
        )
        self.assertEqual(collist, columns)
        self.assertEqual(data, self.data)



#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

from unittest.mock import call

from openstackclient.tests.unit.volume.v2 import fakes as volume_fakes
from openstackclient.volume.v2 import consistency_group_snapshot


class TestConsistencyGroupSnapshot(volume_fakes.TestVolume):
    def setUp(self):
        super().setUp()

        # Get a shortcut to the TransferManager Mock
        self.cgsnapshots_mock = self.volume_client.cgsnapshots
        self.cgsnapshots_mock.reset_mock()
        self.consistencygroups_mock = self.volume_client.consistencygroups
        self.consistencygroups_mock.reset_mock()


class TestConsistencyGroupSnapshotCreate(TestConsistencyGroupSnapshot):
    _consistency_group_snapshot = (
        volume_fakes.create_one_consistency_group_snapshot()
    )
    consistency_group = volume_fakes.create_one_consistency_group()

    columns = (
        'consistencygroup_id',
        'created_at',
        'description',
        'id',
        'name',
        'status',
    )
    data = (
        _consistency_group_snapshot.consistencygroup_id,
        _consistency_group_snapshot.created_at,
        _consistency_group_snapshot.description,
        _consistency_group_snapshot.id,
        _consistency_group_snapshot.name,
        _consistency_group_snapshot.status,
    )

    def setUp(self):
        super().setUp()
        self.cgsnapshots_mock.create.return_value = (
            self._consistency_group_snapshot
        )
        self.consistencygroups_mock.get.return_value = self.consistency_group

        # Get the command object to test
        self.cmd = consistency_group_snapshot.CreateConsistencyGroupSnapshot(
            self.app, None
        )

    def test_consistency_group_snapshot_create(self):
        arglist = [
            '--consistency-group',
            self.consistency_group.id,
            '--description',
            self._consistency_group_snapshot.description,
            self._consistency_group_snapshot.name,
        ]
        verifylist = [
            ('consistency_group', self.consistency_group.id),
            ('description', self._consistency_group_snapshot.description),
            ('snapshot_name', self._consistency_group_snapshot.name),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.consistencygroups_mock.get.assert_called_once_with(
            self.consistency_group.id
        )
        self.cgsnapshots_mock.create.assert_called_once_with(
            self.consistency_group.id,
            name=self._consistency_group_snapshot.name,
            description=self._consistency_group_snapshot.description,
        )

        self.assertEqual(self.columns, columns)
        self.assertEqual(self.data, data)

    def test_consistency_group_snapshot_create_no_consistency_group(self):
        arglist = [
            '--description',
            self._consistency_group_snapshot.description,
            self._consistency_group_snapshot.name,
        ]
        verifylist = [
            ('description', self._consistency_group_snapshot.description),
            ('snapshot_name', self._consistency_group_snapshot.name),
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        columns, data = self.cmd.take_action(parsed_args)

        self.consistencygroups_mock.get.assert_called_once_with(
            self._consistency_group_snapshot.name
        )
        self.cgsnapshots_mock.create.assert_called_once_with(
            self.consistency_group.id,
            name=self._consistency_group_snapshot.name,
            description=self._consistency_group_snapshot.description,
        )

        self.assertEqual(self.columns, columns)
        self.assertEqual(self.data, data)


class TestConsistencyGroupSnapshotDelete(TestConsistencyGroupSnapshot):
    consistency_group_snapshots = (
        volume_fakes.create_consistency_group_snapshots(count=2)
    )

    def setUp(self):
        super().setUp()

        self.cgsnapshots_mock.get = (
            volume_fakes.get_consistency_group_snapshots(
                self.consistency_group_snapshots
            )
        )
        self.cgsnapshots_mock.delete.return_value = None

        # Get the command object to mock
        self.cmd = consistency_group_snapshot.DeleteConsistencyGroupSnapshot(
            self.app, None
        )

    def test_consistency_group_snapshot_delete(self):
        arglist = [self.consistency_group_snapshots[0].id]
        verifylist = [
            (
                "consistency_group_snapshot",
                [self.consistency_group_snapshots[0].id],
            )
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)

        result = self.cmd.take_action(parsed_args)

        self.cgsnapshots_mock.delete.assert_called_once_with(
            self.consistency_group_snapshots[0].id
        )
        self.assertIsNone(result)

    def test_multiple_consistency_group_snapshots_delete(self):
        arglist = []
        for c in self.consistency_group_snapshots:
            arglist.append(c.id)
        verifylist = [
            ('consistency_group_snapshot', arglist),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        result = self.cmd.take_action(parsed_args)

        calls = []
        for c in self.consistency_group_snapshots:
            calls.append(call(c.id))
        self.cgsnapshots_mock.delete.assert_has_calls(calls)
        self.assertIsNone(result)


class TestConsistencyGroupSnapshotList(TestConsistencyGroupSnapshot):
    consistency_group_snapshots = (
        volume_fakes.create_consistency_group_snapshots(count=2)
    )
    consistency_group = volume_fakes.create_one_consistency_group()

    columns = [
        'ID',
        'Status',
        'Name',
    ]
    columns_long = [
        'ID',
        'Status',
        'ConsistencyGroup ID',
        'Name',
        'Description',
        'Created At',
    ]
    data = []
    for c in consistency_group_snapshots:
        data.append(
            (
                c.id,
                c.status,
                c.name,
            )
        )
    data_long = []
    for c in consistency_group_snapshots:
        data_long.append(
            (
                c.id,
                c.status,
                c.consistencygroup_id,
                c.name,
                c.description,
                c.created_at,
            )
        )

    def setUp(self):
        super().setUp()

        self.cgsnapshots_mock.list.return_value = (
            self.consistency_group_snapshots
        )
        self.consistencygroups_mock.get.return_value = self.consistency_group
        # Get the command to test
        self.cmd = consistency_group_snapshot.ListConsistencyGroupSnapshot(
            self.app, None
        )

    def test_consistency_group_snapshot_list_without_options(self):
        arglist = []
        verifylist = [
            ("all_projects", False),
            ("long", False),
            ("status", None),
            ("consistency_group", None),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)

        search_opts = {
            'all_tenants': False,
            'status': None,
            'consistencygroup_id': None,
        }
        self.cgsnapshots_mock.list.assert_called_once_with(
            detailed=True, search_opts=search_opts
        )
        self.assertEqual(self.columns, columns)
        self.assertEqual(self.data, list(data))

    def test_consistency_group_snapshot_list_with_long(self):
        arglist = [
            "--long",
        ]
        verifylist = [
            ("all_projects", False),
            ("long", True),
            ("status", None),
            ("consistency_group", None),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)

        search_opts = {
            'all_tenants': False,
            'status': None,
            'consistencygroup_id': None,
        }
        self.cgsnapshots_mock.list.assert_called_once_with(
            detailed=True, search_opts=search_opts
        )
        self.assertEqual(self.columns_long, columns)
        self.assertEqual(self.data_long, list(data))

    def test_consistency_group_snapshot_list_with_options(self):
        arglist = [
            "--all-project",
            "--status",
            self.consistency_group_snapshots[0].status,
            "--consistency-group",
            self.consistency_group.id,
        ]
        verifylist = [
            ("all_projects", True),
            ("long", False),
            ("status", self.consistency_group_snapshots[0].status),
            ("consistency_group", self.consistency_group.id),
        ]

        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)

        search_opts = {
            'all_tenants': True,
            'status': self.consistency_group_snapshots[0].status,
            'consistencygroup_id': self.consistency_group.id,
        }
        self.consistencygroups_mock.get.assert_called_once_with(
            self.consistency_group.id
        )
        self.cgsnapshots_mock.list.assert_called_once_with(
            detailed=True, search_opts=search_opts
        )
        self.assertEqual(self.columns, columns)
        self.assertEqual(self.data, list(data))


class TestConsistencyGroupSnapshotShow(TestConsistencyGroupSnapshot):
    _consistency_group_snapshot = (
        volume_fakes.create_one_consistency_group_snapshot()
    )

    columns = (
        'consistencygroup_id',
        'created_at',
        'description',
        'id',
        'name',
        'status',
    )
    data = (
        _consistency_group_snapshot.consistencygroup_id,
        _consistency_group_snapshot.created_at,
        _consistency_group_snapshot.description,
        _consistency_group_snapshot.id,
        _consistency_group_snapshot.name,
        _consistency_group_snapshot.status,
    )

    def setUp(self):
        super().setUp()

        self.cgsnapshots_mock.get.return_value = (
            self._consistency_group_snapshot
        )
        self.cmd = consistency_group_snapshot.ShowConsistencyGroupSnapshot(
            self.app, None
        )

    def test_consistency_group_snapshot_show(self):
        arglist = [self._consistency_group_snapshot.id]
        verifylist = [
            ("consistency_group_snapshot", self._consistency_group_snapshot.id)
        ]
        parsed_args = self.check_parser(self.cmd, arglist, verifylist)
        columns, data = self.cmd.take_action(parsed_args)
        self.cgsnapshots_mock.get.assert_called_once_with(
            self._consistency_group_snapshot.id
        )
        self.assertEqual(self.columns, columns)
        self.assertEqual(self.data, data)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import importlib.util
import os
import re
import subprocess
import sys
import unittest

import fixtures

ROOT_DIR = os.path.normpath(
    os.path.join(os.path.dirname(__file__), '..', '..', '..')
)
SELFTEST_REGEX = re.compile(r'\b(Okay|[HEW]\d{3}|O\d{3}):\s(.*)')

# Checks that filter on 'openstackclient/tests/unit' in the filename need the
# temp file written into that path structure so the check is not skipped.
_UNIT_TEST_SUBDIRS = {
    'O401': os.path.join('openstackclient', 'tests', 'unit'),
    'O402': os.path.join('openstackclient', 'tests', 'unit'),
}


def _load_checks():
    spec = importlib.util.spec_from_file_location(
        '_osc_hacking_checks',
        os.path.join(ROOT_DIR, 'hacking', 'checks.py'),
    )
    module = importlib.util.module_from_spec(spec)
    spec.loader.exec_module(module)
    return module


def _get_examples(check):
    for line in check.__doc__.splitlines():
        line = line.lstrip()
        match = SELFTEST_REGEX.match(line)
        if match:
            yield match.group(1), match.group(2)


class HackingTestCase(unittest.TestCase):
    def _test_check(self, code, source):
        lines = [
            part.replace(r'\t', '\t') + '\n' for part in source.split(r'\n')
        ]
        subdir = {
            'O401': os.path.join('openstackclient', 'tests', 'unit'),
            'O402': os.path.join('openstackclient', 'tests', 'unit'),
        }.get(code, '')

        with fixtures.TempDir() as tmp:
            dirpath = os.path.join(tmp.path, subdir) if subdir else tmp.path
            if subdir:
                os.makedirs(dirpath)

            fpath = os.path.join(dirpath, 'test_tmp.py')
            with open(fpath, 'w') as f:
                f.write(''.join(lines))

            cmd = [
                sys.executable,
                '-mflake8',
                '--config',
                os.path.join(ROOT_DIR, 'tox.ini'),
                f'--select={code}',
                '--format=%(code)s\t%(path)s\t%(row)d',
                fpath,
            ]
            out, _ = subprocess.Popen(
                cmd, stdout=subprocess.PIPE, cwd=ROOT_DIR
            ).communicate()
            out = out.decode('utf-8')

            if code == 'Okay':
                self.assertEqual('', out)
            else:
                self.assertNotEqual('', out, f"Failed to trigger rule {code}")
                self.assertEqual(code, out.split('\t')[0].rstrip(':'), out)

    def test_checks(self):
        checks_module = _load_checks()

        for name in sorted(dir(checks_module)):
            check = getattr(checks_module, name)
            if not callable(check):
                continue

            if getattr(check, 'skip_on_py3', None) is not False:
                continue

            if not check.__doc__:
                continue

            for code, source in _get_examples(check):
                with self.subTest(check=name, example=source):
                    self._test_check(code, source)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import uuid

from openstackclient.tests.functional import base


class ServerGroupTests(base.TestCase):
    """Functional tests for servergroup."""

    def test_server_group_delete(self):
        """Test create, delete multiple"""
        name1 = uuid.uuid4().hex
        name2 = uuid.uuid4().hex
        cmd_output = self.openstack(
            'server group create ' + '--policy affinity ' + name1,
            parse_output=True,
        )
        self.assertEqual(name1, cmd_output['name'])
        self.assertEqual('affinity', cmd_output['policy'])

        cmd_output = self.openstack(
            'server group create ' + '--policy anti-affinity ' + name2,
            parse_output=True,
        )
        self.assertEqual(name2, cmd_output['name'])
        self.assertEqual('anti-affinity', cmd_output['policy'])

        del_output = self.openstack(
            'server group delete ' + name1 + ' ' + name2
        )
        self.assertOutput('', del_output)

    def test_server_group_show_and_list(self):
        """Test server group create, show, and list"""
        name1 = uuid.uuid4().hex
        name2 = uuid.uuid4().hex

        # test server group show
        cmd_output = self.openstack(
            'server group create ' + '--policy affinity ' + name1,
            parse_output=True,
        )
        self.addCleanup(self.openstack, 'server group delete ' + name1)
        cmd_output = self.openstack(
            'server group show ' + name1,
            parse_output=True,
        )
        self.assertEqual(name1, cmd_output['name'])
        self.assertEqual('affinity', cmd_output['policy'])

        cmd_output = self.openstack(
            'server group create ' + '--policy anti-affinity ' + name2,
            parse_output=True,
        )
        self.addCleanup(self.openstack, 'server group delete ' + name2)
        cmd_output = self.openstack(
            'server group show ' + name2,
            parse_output=True,
        )
        self.assertEqual(name2, cmd_output['name'])
        self.assertEqual('anti-affinity', cmd_output['policy'])

        # test server group list
        cmd_output = self.openstack(
            'server group list',
            parse_output=True,
        )
        names = [x["Name"] for x in cmd_output]
        self.assertIn(name1, names)
        self.assertIn(name2, names)
        policies = [x["Policy"] for x in cmd_output]
        self.assertIn('affinity', policies)
        self.assertIn('anti-affinity', policies)



#   Copyright 2012-2013 OpenStack Foundation
#
#   Licensed under the Apache License, Version 2.0 (the "License"); you may
#   not use this file except in compliance with the License. You may obtain
#   a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#   WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#   License for the specific language governing permissions and limitations
#   under the License.
#

"""Compute v2 Console action implementations"""

import argparse
from collections.abc import Iterable, Sequence
from typing import Any

from osc_lib.cli import parseractions
from osc_lib import utils

from openstackclient import command
from openstackclient.i18n import _


def _get_console_columns(item: Any) -> tuple[tuple[str, ...], tuple[str, ...]]:
    # To maintain backwards compatibility we need to rename sdk props to
    # whatever OSC was using before
    hidden_columns = ['id', 'links', 'location', 'name']
    return utils.get_osc_show_columns_for_sdk_resource(
        item, {}, hidden_columns
    )


class ShowConsoleLog(command.Command):
    _description = _("Show server's console output")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'server',
            metavar='<server>',
            help=_("Server to show console log (name or ID)"),
        )
        parser.add_argument(
            '--lines',
            metavar='<num-lines>',
            type=int,
            default=None,
            action=parseractions.NonNegativeAction,
            help=_(
                "Number of lines to display from the end of the log "
                "(default=all)"
            ),
        )
        return parser

    def take_action(self, parsed_args: argparse.Namespace) -> None:
        compute_client = self.app.client_manager.compute

        server = compute_client.find_server(
            name_or_id=parsed_args.server, ignore_missing=False
        )

        output = compute_client.get_server_console_output(
            server.id, length=parsed_args.lines
        )
        data: str | None = None
        if output:
            data = output.get('output', None)

        if data and data[-1] != '\n':
            data += '\n'

        if data:
            self.app.stdout.write(data)


class ShowConsoleURL(command.ShowOne):
    _description = _("Show server's remote console URL")

    def get_parser(self, prog_name: str) -> argparse.ArgumentParser:
        parser = super().get_parser(prog_name)
        parser.add_argument(
            'server',
            metavar='<server>',
            help=_("Server to show URL (name or ID)"),
        )
        type_group = parser.add_mutually_exclusive_group()
        type_group.add_argument(
            '--novnc',
            dest='url_type',
            action='store_const',
            const='novnc',
            default='novnc',
            help=_("Show noVNC console URL (default)"),
        )
        type_group.add_argument(
            '--xvpvnc',
            dest='url_type',
            action='store_const',
            const='xvpvnc',
            help=_("Show xvpvnc console URL"),
        )
        type_group.add_argument(
            '--spice',
            dest='url_type',
            action='store_const',
            const='spice-html5',
            help=_("Show SPICE console URL"),
        )
        type_group.add_argument(
            '--spice-direct',
            dest='url_type',
            action='store_const',
            const='spice-direct',
            help=_("Show SPICE direct protocol native console URL"),
        )
        type_group.add_argument(
            '--rdp',
            dest='url_type',
            action='store_const',
            const='rdp-html5',
            help=_("Show RDP console URL"),
        )
        type_group.add_argument(
            '--serial',
            dest='url_type',
            action='store_const',
            const='serial',
            help=_("Show serial console URL"),
        )
        type_group.add_argument(
            '--mks',
            dest='url_type',
            action='store_const',
            const='webmks',
            help=_("Show WebMKS console URL"),
        )
        return parser

    def take_action(
        self, parsed_args: argparse.Namespace
    ) -> tuple[Sequence[str], Iterable[Any]]:
        compute_client = self.app.client_manager.compute
        server = compute_client.find_server(
            parsed_args.server, ignore_missing=False
        )

        console = compute_client.create_console(
            server.id, console_type=parsed_args.url_type
        )

        display_columns, columns = _get_console_columns(console)
        data = utils.get_dict_properties(console, columns)

        return (display_columns, data)



# Copyright (c) 2016, Intel Corporation.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import uuid

from openstackclient.tests.functional.network.v2 import common


class TestMeter(common.NetworkTests):
    """Functional tests for network meter"""

    def setUp(self):
        super().setUp()

        if not self.is_extension_enabled("metering"):
            self.skipTest("No metering extension present")

    # NOTE(dtroyer): Do not normalize the setup and teardown of the resource
    #                creation and deletion.  Little is gained when each test
    #                has its own needs and there are collisions when running
    #                tests in parallel.

    def test_meter_delete(self):
        """Test create, delete multiple"""
        name1 = uuid.uuid4().hex
        name2 = uuid.uuid4().hex
        description = 'fakedescription'
        json_output = self.openstack(
            'network meter create '
            + ' --description '
            + description
            + ' '
            + name1,
            parse_output=True,
        )
        self.assertEqual(
            name1,
            json_output.get('name'),
        )
        # Check if default shared values
        self.assertFalse(json_output.get('shared'))
        self.assertEqual(
            'fakedescription',
            json_output.get('description'),
        )

        json_output_2 = self.openstack(
            'network meter create '
            + '--description '
            + description
            + ' '
            + name2,
            parse_output=True,
        )
        self.assertEqual(
            name2,
            json_output_2.get('name'),
        )
        # Check if default shared values
        self.assertFalse(json_output_2.get('shared'))
        self.assertEqual(
            'fakedescription',
            json_output_2.get('description'),
        )

        raw_output = self.openstack(
            'network meter delete ' + name1 + ' ' + name2,
        )
        self.assertOutput('', raw_output)

    def test_meter_list(self):
        """Test create, list filters, delete"""
        name1 = uuid.uuid4().hex
        json_output = self.openstack(
            'network meter create '
            + '--description Test1 '
            + '--share '
            + name1,
            parse_output=True,
        )
        self.addCleanup(self.openstack, 'network meter delete ' + name1)

        self.assertEqual(
            'Test1',
            json_output.get('description'),
        )
        self.assertTrue(json_output.get('shared'))

        name2 = uuid.uuid4().hex
        json_output_2 = self.openstack(
            'network meter create '
            + '--description Test2 '
            + '--no-share '
            + name2,
            parse_output=True,
        )
        self.addCleanup(self.openstack, 'network meter delete ' + name2)

        self.assertEqual(
            'Test2',
            json_output_2.get('description'),
        )
        self.assertFalse(
            json_output_2.get('shared'),
        )

        raw_output = self.openstack('network meter list', parse_output=True)
        name_list = [item.get('Name') for item in raw_output]
        self.assertIn(name1, name_list)
        self.assertIn(name2, name_list)

    def test_meter_show(self):
        """Test create, show, delete"""
        name1 = uuid.uuid4().hex
        description = 'fakedescription'
        json_output = self.openstack(
            'network meter create '
            + ' --description '
            + description
            + ' '
            + name1,
            parse_output=True,
        )
        meter_id = json_output.get('id')
        self.addCleanup(self.openstack, 'network meter delete ' + name1)

        # Test show with ID
        json_output = self.openstack(
            'network meter show ' + meter_id,
            parse_output=True,
        )
        self.assertFalse(json_output.get('shared'))
        self.assertEqual(
            'fakedescription',
            json_output.get('description'),
        )
        self.assertEqual(
            name1,
            json_output.get('name'),
        )

        # Test show with name
        json_output = self.openstack(
            'network meter show ' + name1,
            parse_output=True,
        )
        self.assertEqual(
            meter_id,
            json_output.get('id'),
        )
        self.assertFalse(json_output.get('shared'))
        self.assertEqual(
            'fakedescription',
            json_output.get('description'),
        )
